import inspect

import pytest


def _is_iivsearch(obj):
    params = inspect.signature(obj).parameters
    assert 'algorithm' in params.keys()
    assert 'rank_type' in params.keys()
    assert params['rank_type'].default == 'mbic'
    assert 'kwargs' in params.keys()
    assert 'kwargs' in obj.__doc__
    return True


def _is_modelsearch(obj):
    params = inspect.signature(obj).parameters
    assert 'search_space' in params.keys()
    assert 'rank_type' in params.keys()
    assert params['rank_type'].default == 'mbic'
    assert 'kwargs' in params.keys()
    assert 'kwargs' in obj.__doc__
    return True


def _dynamic_tool_import_1(tool: str):
    mod = __import__('pharmpy.tools', globals(), locals(), [tool], 0)
    return getattr(mod, tool)


def _dynamic_tool_import_2(tool: str):
    import pharmpy.tools as mod

    return getattr(mod, tool)


@pytest.mark.parametrize(
    ('tool_import',),
    (
        (_dynamic_tool_import_1,),
        (_dynamic_tool_import_2,),
    ),
)
@pytest.mark.parametrize(
    ('tool', 'identity'),
    (
        ('run_iivsearch', _is_iivsearch),
        ('run_modelsearch', _is_modelsearch),
    ),
)
def test_import_tools_getattr(tool_import, tool, identity):
    obj = tool_import(tool)
    assert identity(obj)


def test_import_tools_import_run_iivsearch():
    from pharmpy.tools import run_iivsearch

    assert _is_iivsearch(run_iivsearch)


def test_import_tools_import_run_modelsearch():
    from pharmpy.tools import run_modelsearch

    assert _is_modelsearch(run_modelsearch)



from typing import Dict, List, Optional, Union

from pharmpy.deps import pandas as pd
from pharmpy.model import Model, Parameter, Parameters, RandomVariables


def get_thetas(model: Model):
    """Get all thetas (structural parameters) of a model

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Parameters
        A copy of all theta parameters

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> get_thetas(model)
                value  lower upper    fix
    PTVCL    0.004693   0.00     ∞  False
    PTVV     1.009160   0.00     ∞  False
    THETA_3  0.100000  -0.99     ∞  False

    See also
    --------
    get_omegas : Get omega parameters
    get_sigmas : Get sigma parameters
    """
    rvs_fs = model.random_variables.free_symbols
    thetas = [p for p in model.parameters if p.symbol not in rvs_fs]
    return Parameters(tuple(thetas))


def get_omegas(model: Model):
    """Get all omegas (variability parameters) of a model

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Parameters
        A copy of all omega parameters

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> get_omegas(model)
             value  lower upper    fix
    IVCL  0.030963    0.0     ∞  False
    IVV   0.031128    0.0     ∞  False

    See also
    --------
    get_thetas : Get theta parameters
    get_sigmas : Get sigma parameters
    """
    omegas = [p for p in model.parameters if p.symbol in model.random_variables.etas.free_symbols]
    return Parameters(tuple(omegas))


def get_sigmas(model: Model):
    """Get all sigmas (residual error variability parameters) of a model

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Parameters
        A copy of all sigma parameters

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> get_sigmas(model)
                  value  lower upper    fix
    SIGMA_1_1  0.013241    0.0     ∞  False

    See also
    --------
    get_thetas : Get theta parameters
    get_omegas : Get omega parameters
    """
    sigmas = [
        p for p in model.parameters if p.symbol in model.random_variables.epsilons.free_symbols
    ]
    return Parameters(tuple(sigmas))


def set_initial_estimates(model: Model, inits: Dict[str, float]):
    """Set initial estimates

    Parameters
    ----------
    model : Model
        Pharmpy model
    inits : dict
        A dictionary of parameter init for parameters to change

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_initial_estimates
    >>> model = load_example_model("pheno")
    >>> model = set_initial_estimates(model, {'PTVCL': 2})
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 2.0, lower=0.0, upper=∞, fix=False)

    See also
    --------
    fix_parameters_to : Fixing and setting parameter initial estimates in the same function
    unfix_paramaters_to : Unfixing parameters and setting a new initial estimate in the same
    """
    new = model.parameters.set_initial_estimates(inits)
    model = model.replace(parameters=new)
    return model.update_source()


def update_inits(
    model: Model, parameter_estimates: pd.Series, move_est_close_to_bounds: bool = False
):
    """Update initial parameter estimate for a model

    Updates initial estimates of population parameters for a model.
    If the new initial estimates are out of bounds or NaN this function will raise.

    Parameters
    ----------
    model : Model
        Pharmpy model to update initial estimates
    parameter_estimates : pd.Series
        Parameter estimates to update
    move_est_close_to_bounds : bool
        Move estimates that are close to bounds. If correlation >0.99 the correlation will
        be set to 0.9, if variance is <0.001 the variance will be set to 0.01.

    Returns
    -------
    Model
        Pharmpy model object

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, update_inits
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> model.parameters.inits  # doctest:+ELLIPSIS
    {'PTVCL': 0.00469307, 'PTVV': 1.00916, 'THETA_3': 0.1, 'IVCL': 0.0309626, 'IVV': 0.031128, 'SIGMA_1_1': 0.013241}
    >>> model = update_inits(model, results.parameter_estimates)
    >>> model.parameters.inits  # doctest:+ELLIPSIS
    {'PTVCL': 0.00469555, 'PTVV': 0.984258, 'THETA_3': 0.15892, 'IVCL': 0.0293508, 'IVV': 0.027906, ...}

    """
    # FIXME: Can be combined with set_initial_estimates
    if move_est_close_to_bounds:
        parameter_estimates = _move_est_close_to_bounds(model, parameter_estimates)

    model = model.replace(parameters=model.parameters.set_initial_estimates(parameter_estimates))

    return model.update_source()


def _get_nonfixed_rvs(model):
    fixed_omegas = get_omegas(model).fixed.names
    rvs = model.random_variables
    nonfixed_rvs = [rv for rv in rvs if str(list(rv.variance.free_symbols)[0]) not in fixed_omegas]
    return RandomVariables.create(nonfixed_rvs)


def _move_est_close_to_bounds(model: Model, pe):
    rvs, pset = _get_nonfixed_rvs(model), model.parameters
    est = pe.to_dict()
    sdcorr = rvs.parameters_sdcorr(est)
    newdict = est.copy()
    for dist in rvs:
        rvs = dist.names
        if len(rvs) > 1:
            sigma_sym = dist.variance
            for i in range(sigma_sym.rows):
                for j in range(sigma_sym.cols):
                    param_name = sigma_sym[i, j].name
                    if i != j:
                        if sdcorr[param_name] > 0.99:
                            name_i, name_j = sigma_sym[i, i].name, sigma_sym[j, j].name
                            # From correlation to covariance
                            corr_new = 0.9
                            sd_i, sd_j = sdcorr[name_i], sdcorr[name_j]
                            newdict[param_name] = corr_new * sd_i * sd_j
                    else:
                        if not _is_zero_fix(pset[param_name]) and est[param_name] < 0.001:
                            newdict[param_name] = 0.01
        else:
            param_name = dist.variance.name
            if not pset[param_name].fix:
                if not _is_zero_fix(pset[param_name]) and est[param_name] < 0.001:
                    newdict[param_name] = 0.01
    return newdict


def _is_zero_fix(param):
    return param.init == 0 and param.fix


def set_upper_bounds(model: Model, bounds: Dict[str, float]):
    """Set parameter upper bounds

    Parameters
    ----------
    model : Model
        Pharmpy model
    bounds : dict
        A dictionary of parameter bounds for parameters to change

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_upper_bounds
    >>> model = load_example_model("pheno")
    >>> model = set_upper_bounds(model, {'PTVCL': 10})
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=0.0, upper=10, fix=False)

    See also
    --------
    set_lower_bounds : Set parameter lower bounds
    unconstrain_parameters : Remove all constraints of parameters
    """
    new = []
    for p in model.parameters:
        if p.name in bounds:
            newparam = Parameter(
                name=p.name, init=p.init, lower=p.lower, upper=bounds[p.name], fix=p.fix
            )
        else:
            newparam = p
        new.append(newparam)
    model = model.replace(parameters=Parameters(tuple(new)))
    return model.update_source()


def set_lower_bounds(model: Model, bounds: Dict[str, float]):
    """Set parameter lower bounds

    Parameters
    ----------
    model : Model
        Pharmpy model
    bounds : dict
        A dictionary of parameter bounds for parameters to change

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_lower_bounds
    >>> model = load_example_model("pheno")
    >>> model = set_lower_bounds(model, {'PTVCL': -10})
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=-10, upper=∞, fix=False)

    See also
    --------
    set_upper_bounds : Set parameter upper bounds
    unconstrain_parameters : Remove all constraints of parameters
    """
    new = []
    for p in model.parameters:
        if p.name in bounds:
            newparam = Parameter(
                name=p.name, init=p.init, lower=bounds[p.name], upper=p.upper, fix=p.fix
            )
        else:
            newparam = p
        new.append(newparam)
    model = model.replace(parameters=Parameters(tuple(new)))
    model = model.update_source()
    return model


def fix_parameters(model: Model, parameter_names: Union[List[str], str]):
    """Fix parameters

    Fix all listed parameters

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_names : list or str
        one parameter name or a list of parameter names

    Returns
    -------
    Model
        Pharmpy model object

    Example
    -------
    >>> from pharmpy.modeling import fix_parameters, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=0.0, upper=∞, fix=False)
    >>> model = fix_parameters(model, 'PTVCL')
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=0.0, upper=∞, fix=True)

    See also
    --------
    fix_or_unfix_parameters : Fix or unfix parameters (given boolean)
    fix_parameters_to : Fixing and setting parameter initial estimates in the same function
    unfix_paramaters : Unfixing parameters
    unfix_paramaters_to : Unfixing parameters and setting a new initial estimate in the same
        function

    """
    if isinstance(parameter_names, str):
        parameter_names = [parameter_names]
    params = model.parameters
    new = []
    for p in params:
        if p.name in parameter_names:
            new_param = p.replace(fix=True)
        else:
            new_param = p
        new.append(new_param)
    model = model.replace(parameters=Parameters(tuple(new)))
    return model.update_source()


def unfix_parameters(model: Model, parameter_names: Union[List[str], str]):
    """Unfix parameters

    Unfix all listed parameters

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_names : list or str
        one parameter name or a list of parameter names

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import fix_parameters, unfix_parameters, load_example_model
    >>> model = load_example_model("pheno")
    >>> model = fix_parameters(model, ['PTVCL', 'PTVV', 'THETA_3'])
    >>> model.parameters.fix    # doctest: +ELLIPSIS
    {'PTVCL': True, 'PTVV': True, 'THETA_3': True, ...}
    >>> model = unfix_parameters(model, 'PTVCL')
    >>> model.parameters.fix        # doctest: +ELLIPSIS
    {'PTVCL': False, 'PTVV': True, 'THETA_3': True, ...}

    See also
    --------
    unfix_paramaters_to : Unfixing parameters and setting a new initial estimate in the same
        function
    fix_parameters : Fix parameters
    fix_or_unfix_parameters : Fix or unfix parameters (given boolean)
    fix_parameters_to : Fixing and setting parameter initial estimates in the same function
    unconstrain_parameters : Remove all constraints of parameters

    """
    if isinstance(parameter_names, str):
        parameter_names = [parameter_names]
    params = model.parameters
    new = []
    for p in params:
        if p.name in parameter_names:
            new_param = p.replace(fix=False)
        else:
            new_param = p
        new.append(new_param)
    model = model.replace(parameters=Parameters(tuple(new)))
    model = model.update_source()
    return model


def fix_parameters_to(model: Model, inits: Dict[str, float]):
    """Fix parameters to

    Fix all listed parameters to specified value/values

    Parameters
    ----------
    model : Model
        Pharmpy model
    inits : dict
        Inits for all parameters to fix and set init

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import fix_parameters_to, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=0.0, upper=∞, fix=False)
    >>> model = fix_parameters_to(model, {'PTVCL': 0.5})
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.5, lower=0.0, upper=∞, fix=True)

    See also
    --------
    fix_parameters : Fix parameters
    fix_or_unfix_parameters : Fix or unfix parameters (given boolean)
    unfix_paramaters : Unfixing parameters
    unfix_paramaters_to : Unfixing parameters and setting a new initial estimate in the same
        function

    """
    model = fix_parameters(model, inits.keys())
    model = set_initial_estimates(model, inits)
    return model


def unfix_parameters_to(model: Model, inits: Dict[str, float]):
    """Unfix parameters to

    Unfix all listed parameters to specified value/values

    Parameters
    ----------
    model : Model
        Pharmpy model
    inits : dict
        Inits for all parameters to unfix and change init

    Examples
    --------
    >>> from pharmpy.modeling import fix_parameters, unfix_parameters_to, load_example_model
    >>> model = load_example_model("pheno")
    >>> model = fix_parameters(model, ['PTVCL', 'PTVV', 'THETA_3'])
    >>> model.parameters.fix    # doctest: +ELLIPSIS
    {'PTVCL': True, 'PTVV': True, 'THETA_3': True, 'IVCL': False, 'IVV': False, 'SIGMA_1_1': False}
    >>> model = unfix_parameters_to(model, {'PTVCL': 0.5})
    >>> model.parameters.fix        # doctest: +ELLIPSIS
    {'PTVCL': False, 'PTVV': True, 'THETA_3': True, ...}
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.5, lower=0.0, upper=∞, fix=False)

    Returns
    -------
    Model
        Pharmpy model object

    See also
    --------
    fix_parameters : Fix parameters
    fix_or_unfix_parameters : Fix or unfix parameters (given boolean)
    unfix_paramaters : Unfixing parameters
    fix_paramaters_to : Fixing parameters and setting a new initial estimate in the same
        function
    """
    model = unfix_parameters(model, inits.keys())
    model = set_initial_estimates(model, inits)
    return model


def fix_or_unfix_parameters(model: Model, parameters: Dict[str, bool]):
    """Fix or unfix parameters

    Set fixedness of parameters to specified values

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameters : dict
        Set fix/unfix for these parameters

    Examples
    --------
    >>> from pharmpy.modeling import fix_or_unfix_parameters, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=0.0, upper=∞, fix=False)
    >>> model = fix_or_unfix_parameters(model, {'PTVCL': True})
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=0.0, upper=∞, fix=True)

    Returns
    -------
    Model
        Pharmpy model object

    See also
    --------
    fix_parameters : Fix parameters
    unfix_paramaters : Unfixing parameters
    fix_paramaters_to : Fixing parameters and setting a new initial estimate in the same
        function
    unfix_paramaters_to : Unfixing parameters and setting a new initial estimate in the same
        function
    """
    params = model.parameters
    new = []
    for p in params:
        if p.name in parameters:
            new_param = p.replace(fix=parameters[p.name])
        else:
            new_param = p
        new.append(new_param)
    model = model.replace(parameters=Parameters(tuple(new)))
    model = model.update_source()
    return model


def unconstrain_parameters(model: Model, parameter_names: List[str]):
    """Remove all constraints from parameters

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_names : list
        Remove all constraints for the listed parameters

    Examples
    --------
    >>> from pharmpy.modeling import unconstrain_parameters, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=0.0, upper=∞, fix=False)
    >>> model = unconstrain_parameters(model, ['PTVCL'])
    >>> model.parameters['PTVCL']
    Parameter("PTVCL", 0.00469307, lower=-∞, upper=∞, fix=False)

    Returns
    -------
    Model
        Pharmpy model object

    See also
    --------
    set_lower_bounds : Set parameter lower bounds
    set_upper_bounds : Set parameter upper bounds
    unfix_parameters : Unfix parameters
    """
    if isinstance(parameter_names, str):
        parameter_names = [parameter_names]
    new = []
    for p in model.parameters:
        if p.name in parameter_names:
            newparam = Parameter(name=p.name, init=p.init)
        else:
            newparam = p
        new.append(newparam)
    model = model.replace(parameters=Parameters(tuple(new)))
    model = model.update_source()
    return model


def add_population_parameter(
    model: Model,
    name: str,
    init: float,
    lower: Optional[float] = None,
    upper: Optional[float] = None,
    fix: bool = False,
):
    """Add a new population parameter to the model

    Parameters
    ----------
    model : Model
        Pharmpy model
    name : str
        Name of the new parameter
    init : float
        Initial estimate of the new parameter
    lower : float
        Lower bound of the new parameter
    upper : float
        Upper bound of the new parameter
    fix : bool
        Should the new parameter be fixed?

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import add_population_parameter, load_example_model
    >>> model = load_example_model("pheno")
    >>> model = add_population_parameter(model, 'POP_KA', 2)
    >>> model.parameters
                  value lower upper    fix
    PTVCL      0.004693   0.0     ∞  False
    PTVV       1.009160   0.0     ∞  False
    THETA_3    0.100000 -0.99     ∞  False
    IVCL       0.030963   0.0     ∞  False
    IVV        0.031128   0.0     ∞  False
    SIGMA_1_1  0.013241   0.0     ∞  False
    POP_KA     2.000000    -∞     ∞  False
    """

    param = Parameter.create(name, init, lower=lower, upper=upper, fix=fix)
    params = model.parameters + param
    model = model.replace(parameters=params)
    return model.update_source()



import numpy as np
import pandas as pd
import pytest

from pharmpy.modeling import (
    add_time_after_dose,
    bin_observations,
    check_dataset,
    deidentify_data,
    drop_columns,
    drop_dropped_columns,
    expand_additional_doses,
    get_cmt,
    get_concentration_parameters_from_data,
    get_covariate_baselines,
    get_doseid,
    get_doses,
    get_evid,
    get_ids,
    get_mdv,
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
    get_observations,
    list_time_varying_covariates,
    load_dataset,
    remove_loq_data,
    set_dvid,
    set_lloq_data,
    set_reference_values,
    translate_nmtran_time,
    undrop_columns,
    unload_dataset,
)
from pharmpy.modeling.basic_models import _create_default_datainfo


def test_get_ids(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    assert get_ids(model) == list(range(1, 60))


def test_get_doseid(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    doseid = get_doseid(model)
    assert len(doseid) == 744
    assert doseid[0] == 1
    assert doseid[743] == 13

    # Same timepoint for dose and observation
    df = model.dataset.copy()
    df.loc[742, 'TIME'] = df.loc[743, 'TIME']
    model = model.replace(dataset=df)
    doseid = get_doseid(model)
    assert len(doseid) == 744
    assert doseid[743] == 12
    assert doseid[742] == 13


def test_get_number_of_individuals(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    assert get_number_of_individuals(model) == 59


def test_get_observations(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    ser = get_observations(model)
    assert ser.loc[1, 2.0] == 17.3
    assert ser.loc[2, 63.5] == 24.6
    assert len(ser) == 155
    s2 = get_observations(model, keep_index=True)
    assert s2.loc[1] == 17.3
    assert s2.loc[11] == 31.0


def test_number_of_observations(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    assert get_number_of_observations(model) == 155
    assert list(get_number_of_observations_per_individual(model)) == [
        2,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        4,
        3,
        1,
        3,
        2,
        4,
        2,
        3,
        3,
        4,
        3,
        3,
        3,
        2,
        3,
        3,
        6,
        2,
        2,
        1,
        1,
        2,
        1,
        3,
        2,
        2,
        2,
        3,
        2,
        4,
        3,
        2,
        3,
        2,
        1,
        3,
        3,
        1,
        1,
        5,
        3,
        4,
        3,
        3,
        2,
        4,
        1,
        1,
        2,
        3,
        3,
    ]


def test_covariate_baselines(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    covs = model.datainfo[['WGT', 'APGR']].set_types('covariate')
    model = model.replace(datainfo=model.datainfo[0:3] + covs + model.datainfo[5:])
    df = get_covariate_baselines(model)
    assert len(df) == 59
    assert list(df.columns) == ['WGT', 'APGR']
    assert df.index.name == 'ID'
    assert df['WGT'].loc[2] == 1.5
    assert df['APGR'].loc[11] == 7.0


def test_doses(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    ser = get_doses(model)
    assert len(ser) == 589
    assert ser.loc[1, 0.0] == 25.0


def test_timevarying_covariates(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    a = list_time_varying_covariates(model)
    assert a == []


def test_get_mdv(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    mdv = get_mdv(model)
    label_test = model.datainfo.typeix['dose'][0].name
    data_test = model.dataset[label_test].astype('float64').squeeze()
    mdv_test = data_test.where(data_test == 0, other=1).astype('int32')
    result = mdv.equals(other=mdv_test)
    assert result is True


def test_get_evid(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    evid = get_evid(model)
    assert evid.sum() == 589


def test_get_cmt(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    cmt = get_cmt(model)
    assert cmt.sum() == 589


def test_add_time_after_dose(load_model_for_test, load_example_model_for_test, testdata):
    m = load_example_model_for_test("pheno")
    m = add_time_after_dose(m)
    tad = m.dataset['TAD']

    assert tad[0] == 0.0
    assert tad[1] == 2.0
    assert tad[743] == 2.0

    m = load_model_for_test(testdata / 'nonmem' / 'models' / 'pef.mod')
    m = add_time_after_dose(m)
    tad = list(m.dataset['TAD'].iloc[0:21])
    assert tad == [
        0.0,
        0.0,
        0.0,
        1.5,
        3.0,
        10.719999999999999,
        0.0,
        0.0,
        0.0,
        1.4500000000000028,
        3.0,
        10.980000000000004,
        0.0,
        0.0,
        2.25,
        3.770000000000003,
        12.0,
        0.0,
        0.0,
        1.4700000000000273,
        2.9700000000000273,
    ]
    assert m.dataset.loc[103, 'TAD'] == 0.0
    assert m.dataset.loc[104, 'TAD'] == pytest.approx(1.17)

    m = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox1.mod')
    m = add_time_after_dose(m)
    tad = list(m.dataset['TAD'].iloc[0:16])
    assert tad == [0.0, 1.0, 1.5, 2.0, 4.0, 6.0, 8.0, 0.0, 12.0, 0.5, 1.0, 1.5, 2.0, 4.0, 6.0, 8.0]


def test_get_concentration_parameters_from_data(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    df = get_concentration_parameters_from_data(model)
    assert df['Cmax'].loc[1, 1] == 17.3


def test_drop_columns(load_example_model_for_test):
    m = load_example_model_for_test('pheno')
    m = drop_columns(m, "APGR")
    correct = ['ID', 'TIME', 'AMT', 'WGT', 'DV', 'FA1', 'FA2']
    assert m.datainfo.names == correct
    assert list(m.dataset.columns) == correct
    m = drop_columns(m, ['DV', 'ID'])
    assert m.datainfo.names == ['TIME', 'AMT', 'WGT', 'FA1', 'FA2']
    assert list(m.dataset.columns) == ['TIME', 'AMT', 'WGT', 'FA1', 'FA2']
    m = drop_columns(m, ['TIME'], mark=True)
    assert m.datainfo['TIME'].drop
    assert list(m.dataset.columns) == ['TIME', 'AMT', 'WGT', 'FA1', 'FA2']


def test_drop_dropped_columns(load_example_model_for_test):
    m = load_example_model_for_test('pheno')
    m = drop_dropped_columns(m)
    correct = ['ID', 'TIME', 'AMT', 'WGT', 'APGR', 'DV', 'FA1', 'FA2']
    assert list(m.dataset.columns) == correct
    m = drop_columns(m, ['ID', 'TIME', 'AMT'], mark=True)
    m = drop_dropped_columns(m)
    assert list(m.dataset.columns) == ['WGT', 'APGR', 'DV', 'FA1', 'FA2']


def test_undrop_columns(load_example_model_for_test):
    m = load_example_model_for_test('pheno')
    m = drop_columns(m, ["APGR", "WGT"], mark=True)
    m = undrop_columns(m, "WGT")
    assert not m.datainfo["WGT"].drop
    assert m.datainfo["APGR"].drop


def test_remove_loq_data(load_example_model_for_test):
    m = load_example_model_for_test('pheno')
    m2 = remove_loq_data(m, lloq=10, uloq=40)
    assert len(m2.dataset) == 736

    df = m.dataset.copy()
    df['LQ'] = [1, 1] + [0] * 742
    m3 = m.replace(dataset=df)
    m3 = remove_loq_data(m3, alq="LQ")
    assert len(m3.dataset) == 743

    m4 = remove_loq_data(m3, blq="LQ")
    assert len(m4.dataset) == 743

    df['LLOQ'] = 10
    m5 = m.replace(dataset=df)
    m5 = remove_loq_data(m5, lloq="LLOQ")
    assert len(m5.dataset) == 742


@pytest.mark.usefixtures('load_example_model_for_test')
@pytest.mark.parametrize(
    'd,expected,keep',
    [
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [1, 1, 0, 0, 1, 1, 0, 0],
                'BLQ': [0, 0, 1, 1, 0, 0, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 4, 5],
            0,
        ),
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [1, 1, 0, 0, 1, 1, 0, 0],
                'BLQ': [1, 1, 1, 1, 1, 1, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 4, 5],
            0,
        ),
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [1, 1, 1, 1, 1, 1, 1, 1],
                'BLQ': [1, 1, 1, 1, 1, 1, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 2, 3, 4, 5, 6, 7],
            0,
        ),
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [1, 1, 0, 0, 1, 1, 0, 0],
                'BLQ': [0, 0, 1, 1, 0, 0, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 2, 4, 5, 6],
            1,
        ),
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [1, 1, 0, 0, 1, 1, 0, 0],
                'BLQ': [0, 0, 1, 1, 0, 0, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 2, 3, 4, 5, 6, 7],
            2,
        ),
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [0, 0, 0, 0, 0, 0, 0, 0],
                'BLQ': [1, 0, 0, 0, 1, 1, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 2, 3, 4],
            1,
        ),
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [0, 0, 0, 0, 0, 0, 0, 0],
                'BLQ': [1, 1, 0, 0, 1, 1, 0, 0],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 2, 3, 4, 6, 7],
            1,
        ),
    ],
)
def test_remove_blq(load_example_model_for_test, d, expected, keep):
    m = load_example_model_for_test('pheno')
    df = pd.DataFrame(d)
    m = m.replace(dataset=df, datainfo=_create_default_datainfo(df))
    new = remove_loq_data(m, blq='BLQ', keep=keep)
    assert list(new.dataset['DV']) == expected


@pytest.mark.usefixtures('load_example_model_for_test')
@pytest.mark.parametrize(
    'd,expected,value',
    [
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [1, 1, 0, 0, 1, 1, 0, 0],
                'BLQ': [0, 0, 1, 1, 0, 0, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 0, 0, 4, 5, 0, 0],
            0,
        ),
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [1, 1, 0, 0, 1, 1, 0, 0],
                'BLQ': [1, 1, 1, 1, 1, 1, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 1, 1, 4, 5, 1, 1],
            1,
        ),
        (
            {
                'ID': [1, 1, 1, 1, 2, 2, 2, 2],
                'MDV': [1, 1, 1, 1, 1, 1, 1, 1],
                'BLQ': [1, 1, 1, 1, 1, 1, 1, 1],
                'DV': [0, 1, 2, 3, 4, 5, 6, 7],
            },
            [0, 1, 2, 3, 4, 5, 6, 7],
            0,
        ),
    ],
)
def test_set_lloq_value(load_example_model_for_test, d, expected, value):
    m = load_example_model_for_test('pheno')
    df = pd.DataFrame(d)
    m = m.replace(dataset=df, datainfo=_create_default_datainfo(df))
    new = set_lloq_data(m, value, blq='BLQ')
    assert list(new.dataset['DV']) == expected


def test_check_dataset(load_example_model_for_test):
    m = load_example_model_for_test('pheno')
    check_dataset(m)

    df = check_dataset(m, verbose=True, dataframe=True)
    assert df is not None
    assert df[df['code'] == 'A1']['result'].iloc[0] == 'OK'
    assert df[df['code'] == 'A4']['result'].iloc[0] == 'SKIP'

    df = m.dataset.copy()
    df.loc[743, 'WGT'] = -1
    m = m.replace(dataset=df)
    df = check_dataset(m, verbose=True, dataframe=True)
    assert df is not None
    assert df[df['code'] == 'A3']['result'].iloc[0] == 'FAIL'


def test_nmtran_time(load_example_model_for_test):
    m = load_example_model_for_test("pheno_linear")
    translate_nmtran_time(m)


def test_expand_additional_doses(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pef.mod')
    model = expand_additional_doses(model)
    df = model.dataset
    assert len(df) == 1494
    assert len(df.columns) == 5
    assert df.loc[0, 'AMT'] == 400.0
    assert df.loc[1, 'AMT'] == 400.0
    assert df.loc[2, 'AMT'] == 400.0
    assert df.loc[3, 'AMT'] == 400.0
    assert df.loc[4, 'AMT'] == 200.0
    assert df.loc[0, 'TIME'] == 0.0
    assert df.loc[1, 'TIME'] == 12.0
    assert df.loc[2, 'TIME'] == 24.0
    assert df.loc[3, 'TIME'] == 36.0
    assert df.loc[4, 'TIME'] == 48.0

    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pef.mod')
    model = expand_additional_doses(model, flag=True)
    df = model.dataset
    assert len(df) == 1494
    assert len(df.columns) == 8
    assert not df.loc[0, 'EXPANDED']
    assert df.loc[1, 'EXPANDED']
    assert df.loc[2, 'EXPANDED']
    assert df.loc[3, 'EXPANDED']
    assert not df.loc[4, 'EXPANDED']


def test_deidentify_data():
    np.random.seed(23)

    example = pd.DataFrame(
        {'ID': [1, 1, 2, 2], 'DATE': ["2012-05-25", "2013-04-02", "2011-12-23", "2005-02-28"]}
    )
    df = deidentify_data(example, date_columns=['DATE'])
    correct = pd.to_datetime(
        pd.Series(["1908-05-25", "1909-04-02", "1907-12-23", "1901-02-28"], name='DATE')
    )
    pd.testing.assert_series_equal(df['DATE'], correct)

    example = pd.DataFrame(
        {
            'ID': [1, 1, 2, 2],
            'DATE': ["2012-05-25", "2013-04-02", "2011-12-23", "2005-02-28"],
            'BIRTH': ["1980-07-07", "1980-07-07", "1956-10-12", "1956-10-12"],
        }
    )
    df = deidentify_data(example, date_columns=['DATE', 'BIRTH'])
    correct_date = pd.to_datetime(
        pd.Series(["1959-12-23", "1953-02-28", "1960-05-25", "1961-04-02"], name='DATE')
    )
    correct_birth = pd.to_datetime(
        pd.Series(["1904-10-12", "1904-10-12", "1928-07-07", "1928-07-07"], name='BIRTH')
    )
    pd.testing.assert_series_equal(df['DATE'], correct_date)
    pd.testing.assert_series_equal(df['BIRTH'], correct_birth)


def test_set_dvid(load_example_model_for_test):
    m = load_example_model_for_test('pheno')
    m = set_dvid(m, 'FA1')
    col = m.datainfo['FA1']
    assert col.type == 'dvid'
    assert col.scale == 'nominal'
    assert col.categories == (0, 1)
    m = set_dvid(m, 'FA1')
    assert m.datainfo['FA1'].type == 'dvid'
    m = set_dvid(m, 'FA2')
    assert m.datainfo['FA1'].type == 'unknown'
    assert m.datainfo['FA2'].type == 'dvid'
    with pytest.raises(ValueError):
        set_dvid(m, 'WGT')


def test_set_reference_values(load_example_model_for_test):
    m = load_example_model_for_test('pheno')
    m2 = set_reference_values(m, {'WGT': 0.5, 'AMT': 4})
    df = m2.dataset
    assert list(df['WGT'].unique()) == [0.5]
    assert df['AMT'][0] == 4.0
    assert df['AMT'][1] == 0.0


def test_unload_dataset(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    assert model.dataset is not None
    model = unload_dataset(model)
    assert model.dataset is None


def test_load_dataset(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    model = unload_dataset(model)
    assert model.dataset is None
    model = load_dataset(model)
    assert model.dataset is not None


def test_bin_observations(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    ser = bin_observations(model, method="equal_width", nbins=10)
    assert ser.iloc[0] == 1
    assert ser[267] == 10
    assert ser.iloc[152] == 8
    assert len(ser) == 155



import pytest

from pharmpy.tools import read_modelfit_results
from pharmpy.workflows import ModelEntry


def test_model_entry_init(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno.mod')

    model_entry = ModelEntry(model)
    assert model_entry.model.name == model.name

    # FIXME: Replace res.log when log is no longer part of ModelfitResults
    model_entry_res = ModelEntry(model, modelfit_results=res, log=res.log)
    assert model_entry_res.modelfit_results.name == res.name


def test_model_entry_create(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno.mod')

    model_entry = ModelEntry.create(model)
    assert model_entry.model.name == model.name

    model_entry_res = ModelEntry(model, modelfit_results=res)
    assert model_entry_res.modelfit_results.name == res.name

    # model = model.replace(name='other')
    # with pytest.raises(ValueError):
    #     ModelEntry.create(model, modelfit_results=res)

    model_parent = model.replace(name='parent')
    model_entry_parent = ModelEntry.create(model, parent=model_parent)
    assert model_entry_parent.parent.name == model_parent.name

    with pytest.raises(ValueError):
        ModelEntry.create(model, parent=model)


def test_attach_results(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno.mod')

    model_entry = ModelEntry(model)
    assert model_entry.model.name == model.name

    model_entry_res = model_entry.attach_results(modelfit_results=res)
    assert model_entry_res.modelfit_results.name == res.name

    with pytest.raises(ValueError):
        model_entry_res.attach_results(modelfit_results=res)



import sympy

from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.model import output
from pharmpy.modeling import add_peripheral_compartment, set_michaelis_menten_elimination


def test_des(load_model_for_test, create_model_for_test, testdata):
    code = """$PROBLEM
$INPUT ID TIME DV AMT
$DATA data.csv IGNORE=@
$SUBROUTINES ADVAN13 TOL=6
$MODEL COMP=(DOSE) COMP=(CENTRAL) COMP=(PERIPH) COMP=(LIVER)
$PK
MAT = THETA(1) * EXP(ETA(1))
CL = THETA(2) * EXP(ETA(2))
VC = THETA(3) * EXP(ETA(3))
Q = THETA(4) * EXP(ETA(4))
VP = THETA(5) * EXP(ETA(5))
VM = THETA(6) * EXP(ETA(6))
KM = THETA(7) * EXP(ETA(7))
KLO = THETA(8) * EXP(ETA(8))
VL = THETA(9) * EXP(ETA(9))
KA = 1/MAT
$DES
DADT(1) = -(KA * A(1))
DADT(2) = KA * A(1) + Q/VP * A(3) - (Q/VC * A(2) + A(2)/VC * VM/(KM + A(2)/VC))
DADT(3) = Q/VC * A(2) - Q/VP * A(3)
DADT(4) = A(2)/VC * VM/(KM + A(2)/VC) - KLO * A(4)
$ERROR
CONC = A(2)/VC
Y = CONC + EPS(1)
$ESTIMATION METHOD=COND INTER
$COVARIANCE PRINT=E
$THETA (0, 1, Inf) ; POP_MAT
$THETA (0, 1, Inf) ; POP_CL
$THETA (0, 1, Inf) ; POP_VC
$THETA (0, 1, Inf) ; POP_Q
$THETA (0, 1, Inf) ; POP_VP
$THETA (0, 1, Inf) ; POP_VM
$THETA (0, 1, Inf) ; POP_KM
$THETA (0, 1, Inf) ; POP_KLO
$THETA (0, 1, Inf) ; POP_VL
$OMEGA 0.1; IIV_MAT
$OMEGA 0.1; IIV_CL
$OMEGA 0.1; IIV_VC
$OMEGA 0.1; IIV_Q
$OMEGA 0.1; IIV_VP
$OMEGA 0.1; IIV_VM
$OMEGA 0.1; IIV_KM
$OMEGA 0.1; IIV_KLO
$OMEGA 0.1; IIV_VL
$SIGMA 0.1; RUV_ADD
"""
    pheno = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = create_model_for_test(code)
    model = model.replace(dataset=pheno.dataset)
    cs = model.statements.ode_system
    assert len(cs) == 4


def test_conversion_round_trip(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    model = add_peripheral_compartment(model)
    model = add_peripheral_compartment(model)
    model = set_michaelis_menten_elimination(model)
    odes = model.statements.ode_system
    central = odes.central_compartment
    assert odes.get_flow(central, output) == parse_expr('CLMM*KM/(V1*(KM + A_CENTRAL(t)/V1))')


def test_des_mm(load_example_model_for_test, create_model_for_test):
    model = load_example_model_for_test('pheno')
    model = add_peripheral_compartment(model)
    model = add_peripheral_compartment(model)
    model = set_michaelis_menten_elimination(model)
    code = model.model_code
    dataset = model.dataset
    model = create_model_for_test(code)
    model = model.replace(dataset=dataset)
    cs = model.statements.ode_system
    central = cs.central_compartment
    assert cs.get_flow(central, output) == sympy.simplify(
        parse_expr('CLMM*KM/(V1*(KM + A_CENTRAL(t)/V1))')
    )



"""
:meta private:
"""
from typing import Literal, Mapping, Set, Union

from pharmpy.deps import sympy
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.unicode import bracket
from pharmpy.model import (
    Assignment,
    Bolus,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    EstimationSteps,
    Infusion,
    Model,
    ModelError,
    Parameter,
    Parameters,
    Statements,
    output,
)
from pharmpy.modeling.help_functions import _as_integer

from .common import remove_unused_parameters_and_rvs, rename_symbols
from .data import get_observations
from .expressions import create_symbol, is_real
from .parameters import (
    add_population_parameter,
    fix_parameters,
    fix_parameters_to,
    set_initial_estimates,
    set_upper_bounds,
    unfix_parameters,
)


def get_and_check_odes(model: Model) -> CompartmentalSystem:
    # Get the ode_system from model and raise if not a CompartmentalSystem
    odes = model.statements.ode_system
    if isinstance(odes, CompartmentalSystem):
        return odes
    else:
        raise ValueError(f'Model {model.name} has no ODE system')


def _extract_params_from_symb(
    statements: Statements, symbol_name: str, pset: Parameters
) -> Parameter:
    terms = {
        symb.name
        for symb in statements.before_odes.full_expression(sympy.Symbol(symbol_name)).free_symbols
    }
    theta_name = terms.intersection(pset.names).pop()
    return pset[theta_name]


def _find_noncov_theta(model, paramsymb, full=False):
    # Deterministic function to find the main theta in the expression for parasymb
    # Set full to True if already providing a full expression
    # Find subexpression with as few thetas and covariates as possible
    # Stop if one theta found with no covs.
    if full:
        start_expr = paramsymb
    else:
        start_expr = model.statements.before_odes.full_expression(paramsymb)

    exprs = [start_expr]
    all_popparams = set(model.parameters.symbols)
    all_covs = set(sympy.Symbol(name) for name in model.datainfo.names)

    while exprs:
        nthetas = float("Inf")
        ncovs = float("Inf")
        next_expr = None
        for expr in exprs:
            symbs = expr.free_symbols
            curthetas = symbs & all_popparams
            ncurthetas = len(curthetas)
            curcovs = symbs & all_covs
            ncurcovs = len(curcovs)
            if ncurthetas != 0 and (
                ncurthetas < nthetas or (ncurthetas == nthetas and ncurcovs < ncovs)
            ):
                next_expr = expr
                thetas = curthetas
                nthetas = ncurthetas
                ncovs = ncurcovs

        if next_expr is None:
            break
        else:
            if nthetas == 1 and ncovs == 0:
                return next(iter(thetas))
            else:
                exprs = next_expr.args
    raise ValueError(f"Could not find theta connected to {paramsymb}")


def add_individual_parameter(model: Model, name: str):
    """Add an individual or pk parameter to a model

    Parameters
    ----------
    model : Model
        Pharmpy model
    name : str
        Name of individual/pk parameter

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_individual_parameter(model, "KA")
    >>> model.statements.find_assignment("KA")
    KA = POP_KA

    """
    model, _ = _add_parameter(model, name)
    model = model.update_source()
    return model


def _add_parameter(
    model: Model, name: str, init: float = 0.1, lower: float = 0, upper: Union[float, None] = None
):
    pops = create_symbol(model, f'POP_{name}')
    model = add_population_parameter(model, pops.name, init, lower=lower, upper=upper)
    symb = create_symbol(model, name)
    ass = Assignment(symb, pops)
    model = model.replace(statements=ass + model.statements)
    return model, symb


def set_first_order_elimination(model: Model):
    """Sets elimination to first order

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_first_order_elimination(model)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CL/V→
    └───────┘

    See also
    --------
    set_zero_order_elimination
    set_michaelis_menten_elimination

    """
    if has_first_order_elimination(model):
        pass
    elif has_zero_order_elimination(model) or has_michaelis_menten_elimination(model):
        model = rename_symbols(model, {'POP_CLMM': 'POP_CL', 'IIV_CLMM': 'IIV_CL'})
        ind = model.statements.find_assignment_index('CLMM')
        assert ind is not None
        clmm_assignment = model.statements[ind]
        assert isinstance(clmm_assignment, Assignment)
        cl_ass = Assignment(sympy.Symbol('CL'), clmm_assignment.expression)
        statements = model.statements[0:ind] + cl_ass + model.statements[ind + 1 :]
        odes = statements.ode_system
        assert isinstance(odes, CompartmentalSystem)
        central = odes.central_compartment
        v = sympy.Symbol('V')
        rate = odes.get_flow(central, output)
        if v not in rate.free_symbols:
            # take first parameter that starts with 'V' and is no longer than 2 characters
            v = [
                idx
                for idx in list(map(lambda x: str(x), rate.free_symbols))
                if idx[0] == 'V' and len(idx) <= 2
            ]
            v = sympy.Symbol(v[0])
        cb = CompartmentalSystemBuilder(odes)
        cb.remove_flow(central, output)
        cb.add_flow(central, output, sympy.Symbol('CL') / v)
        statements = statements.before_odes + CompartmentalSystem(cb) + statements.after_odes
        statements = statements.remove_symbol_definitions(
            {sympy.Symbol('KM')}, statements.ode_system
        )
        model = model.replace(statements=statements)
        model = remove_unused_parameters_and_rvs(model)
    elif has_mixed_mm_fo_elimination(model):
        odes = model.statements.ode_system
        assert odes is not None
        central = odes.central_compartment
        v = sympy.Symbol('V')
        rate = odes.get_flow(central, output)
        if v not in rate.free_symbols:
            v = [
                idx
                for idx in list(map(lambda x: str(x), rate.free_symbols))
                if idx[0] == 'V' and len(idx) <= 2
            ]
            v = sympy.Symbol(v[0])
        cb = CompartmentalSystemBuilder(odes)
        cb.remove_flow(central, output)
        cb.add_flow(central, output, sympy.Symbol('CL') / v)
        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        statements = statements.remove_symbol_definitions(
            {sympy.Symbol('KM'), sympy.Symbol('CLMM')}, statements.ode_system
        )
        model = model.replace(statements=statements)
        model = remove_unused_parameters_and_rvs(model)
    return model


def add_bioavailability(model: Model, add_parameter: bool = True, logit_transform: bool = False):
    """Add bioavailability statement for the first dose compartment of the model.
    Can be added as a new parameter or otherwise it will be set to 1. If added as a parameter,
    a logit transformation can also be applied.

    Parameters
    ----------
    model : Model
        Pharmpy model
    add_parameter : bool
        Add new parameter representing bioavailability or not
    logit_transform : bool
        Logit transform the added bioavailability parameter.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_bioavailability(model)

    See also
    --------
    remove_bioavailability

    """
    odes = get_and_check_odes(model)

    dose_comp = odes.dosing_compartments[0]
    bio = dose_comp.bioavailability

    if isinstance(bio, sympy.Number):
        # Bio not defined
        if add_parameter:
            model, bio_symb = _add_parameter(model, 'BIO', init=float(bio), upper=sympy.Number(1))
            if logit_transform:
                model = model.replace(
                    statements=model.statements.reassign(
                        bio_symb, sympy.log(sympy.Symbol("POP_BIO") / (1 - sympy.Symbol("POP_BIO")))
                    )
                )
                f_ass = Assignment.create(sympy.Symbol('F_BIO'), 1 / (1 + sympy.exp(-bio_symb)))
            else:
                f_ass = Assignment.create(sympy.Symbol('F_BIO'), bio_symb)

            new_before_odes = model.statements.before_odes + f_ass

        else:
            # Add as a number
            bio_ass = Assignment.create(sympy.Symbol("BIO"), sympy.Number(1))
            f_ass = Assignment.create(sympy.Symbol("F_BIO"), bio_ass.symbol)
            new_before_odes = bio_ass + model.statements.before_odes + f_ass

        # Add statement to code
        cb = CompartmentalSystemBuilder(odes)
        cb.set_bioavailability(dose_comp, f_ass.symbol)

        model = model.replace(
            statements=(new_before_odes + CompartmentalSystem(cb) + model.statements.after_odes)
        )

    else:
        # BIO already defined, leave it alone?
        pass

    return model.update_source()


def remove_bioavailability(model: Model):
    """Remove bioavailability from the first dose compartment of model.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = remove_bioavailability(model)

    See also
    --------
    set_bioavailability
    """
    odes = get_and_check_odes(model)
    dosing_comp = odes.dosing_compartments[0]
    bio = dosing_comp.bioavailability
    if bio:
        symbols = bio.free_symbols
        cb = CompartmentalSystemBuilder(odes)
        cb.set_bioavailability(dosing_comp, sympy.Integer(1))
        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        statements = statements.remove_symbol_definitions(symbols, statements.ode_system)
        model = model.replace(statements=statements)
        model = remove_unused_parameters_and_rvs(model)
    return model


def set_zero_order_elimination(model: Model):
    """Sets elimination to zero order.

    Initial estimate for KM is set to 1% of smallest observation.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_zero_order_elimination(model)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CLMM*KM/(V*(KM + A_CENTRAL(t)/V))→
    └───────┘

    See also
    --------
    set_first_order_elimination
    set_michaelis_menten_elimination

    """
    if has_zero_order_elimination(model):
        pass
    elif has_michaelis_menten_elimination(model):
        model = fix_parameters(model, 'POP_KM')
    elif has_mixed_mm_fo_elimination(model):
        model = fix_parameters(model, 'POP_KM')
        odes = get_and_check_odes(model)
        central = odes.central_compartment
        rate = odes.get_flow(central, output)
        rate = subs(rate, {'CL': 0})
        cb = CompartmentalSystemBuilder(odes)
        cb.remove_flow(central, output)
        cb.add_flow(central, output, rate)
        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        statements = statements.remove_symbol_definitions(
            {sympy.Symbol('CL')}, statements.ode_system
        )
        model = model.replace(statements=statements)
        model = remove_unused_parameters_and_rvs(model)
    else:
        model = _do_michaelis_menten_elimination(model)
        if model.dataset is not None:
            obs = get_observations(model)
            init = obs.min() / 100  # 1% of smallest observation
        else:
            init = 0.01
        if init < 0:
            init = 0.01
        model = fix_parameters_to(model, {'POP_KM': init})
    return model


def has_michaelis_menten_elimination(model: Model):
    """Check if the model describes Michaelis-Menten elimination

    This function relies on heuristics and will not be able to detect all
    possible ways of coding the Michaelis-Menten elimination.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        True if model has describes Michaelis-Menten elimination

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_michaelis_menten_elimination(model)
    False
    >>> model = set_michaelis_menten_elimination(model)
    >>> has_michaelis_menten_elimination(model)
    True
    """
    odes = get_and_check_odes(model)
    central = odes.central_compartment
    rate = odes.get_flow(central, output)
    is_nonlinear = odes.t in rate.free_symbols
    is_zero_order = 'POP_KM' in model.parameters and model.parameters['POP_KM'].fix
    could_be_mixed = sympy.Symbol('CL') in rate.free_symbols
    return is_nonlinear and not is_zero_order and not could_be_mixed


def has_zero_order_elimination(model: Model):
    """Check if the model describes zero-order elimination

    This function relies on heuristics and will not be able to detect all
    possible ways of coding the zero-order elimination.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        True if model has describes zero order elimination

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_zero_order_elimination(model)
    False
    >>> model = set_zero_order_elimination(model)
    >>> has_zero_order_elimination(model)
    True
    """
    odes = get_and_check_odes(model)
    central = odes.central_compartment
    rate = odes.get_flow(central, output)
    is_nonlinear = odes.t in rate.free_symbols
    is_zero_order = 'POP_KM' in model.parameters and model.parameters['POP_KM'].fix
    could_be_mixed = sympy.Symbol('CL') in rate.free_symbols
    return is_nonlinear and is_zero_order and not could_be_mixed


def has_mixed_mm_fo_elimination(model: Model):
    """Check if the model describes mixed Michaelis-Menten and first order elimination

    This function relies on heuristics and will not be able to detect all
    possible ways of coding the mixed Michalis-Menten and first order elimination.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        True if model has describes Michaelis-Menten elimination

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_mixed_mm_fo_elimination(model)
    False
    >>> model = set_mixed_mm_fo_elimination(model)
    >>> has_mixed_mm_fo_elimination(model)
    True
    """
    odes = get_and_check_odes(model)
    central = odes.central_compartment
    rate = odes.get_flow(central, output)
    is_nonlinear = odes.t in rate.free_symbols
    is_zero_order = 'POP_KM' in model.parameters and model.parameters['POP_KM'].fix
    could_be_mixed = sympy.Symbol('CL') in rate.free_symbols
    return is_nonlinear and not is_zero_order and could_be_mixed


def has_first_order_elimination(model: Model):
    """Check if the model describes first order elimination

    This function relies on heuristics and will not be able to detect all
    possible ways of coding the first order elimination.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        True if model has describes first order elimination

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_first_order_elimination(model)
    True

    """
    odes = get_and_check_odes(model)
    central = odes.central_compartment
    rate = odes.get_flow(central, output)
    is_nonlinear = odes.t in rate.free_symbols
    return not is_nonlinear


def set_michaelis_menten_elimination(model: Model):
    """Sets elimination to Michaelis-Menten.

    Initial estimate for CLMM is set to CL and KM is set to :math:`max(DV)/2`.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_michaelis_menten_elimination(model)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CLMM*KM/(V*(KM + A_CENTRAL(t)/V))→
    └───────┘

    See also
    --------
    set_first_order_elimination
    set_zero_order_elimination

    """
    if has_michaelis_menten_elimination(model):
        pass
    elif has_zero_order_elimination(model):
        model = unfix_parameters(model, 'POP_KM')
    elif has_mixed_mm_fo_elimination(model):
        odes = get_and_check_odes(model)
        central = odes.central_compartment
        rate = odes.get_flow(central, output)
        rate = subs(rate, {'CL': 0})
        cb = CompartmentalSystemBuilder(odes)
        cb.remove_flow(central, output)
        cb.add_flow(central, output, rate)
        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        statements = statements.remove_symbol_definitions(
            {sympy.Symbol('CL')}, statements.ode_system
        )
        model = model.replace(statements=statements)
        model = remove_unused_parameters_and_rvs(model)
    else:
        model = _do_michaelis_menten_elimination(model)
    return model


def set_mixed_mm_fo_elimination(model: Model):
    """Sets elimination to mixed Michaelis-Menten and first order.

    Initial estimate for CLMM is set to CL/2 and KM is set to :math:`max(DV)/2`.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_mixed_mm_fo_elimination(model)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──(CL + CLMM*KM/(KM + A_CENTRAL(t)/V))/V→
    └───────┘

    See also
    --------
    set_first_order_elimination
    set_zero_order_elimination
    set_michaelis_menten_elimination

    """
    if has_mixed_mm_fo_elimination(model):
        pass
    elif has_michaelis_menten_elimination(model) or has_zero_order_elimination(model):
        model = unfix_parameters(model, 'POP_KM')
        odes = get_and_check_odes(model)
        central = odes.central_compartment
        model = add_individual_parameter(model, 'CL')
        rate = odes.get_flow(central, output)
        assert rate is not None
        cb = CompartmentalSystemBuilder(odes)
        cb.remove_flow(central, output)
        v = sympy.Symbol('V')
        if v not in rate.free_symbols:
            v = sympy.Symbol('VC')
        cb.add_flow(central, output, sympy.Symbol('CL') / v + rate)
        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        model = model.replace(statements=statements)
        model = model.update_source()
    else:
        model = _do_michaelis_menten_elimination(model, combined=True)
    return model


def _do_michaelis_menten_elimination(model: Model, combined: bool = False):
    sset = model.statements
    odes = sset.ode_system
    assert isinstance(odes, CompartmentalSystem)
    central = odes.central_compartment
    old_rate = odes.get_flow(central, output)
    assert old_rate is not None
    numer, denom = old_rate.as_numer_denom()

    km_init, clmm_init = _get_mm_inits(model, numer, combined)

    model, km = _add_parameter(model, 'KM', init=km_init)
    if model.dataset is not None and 'idv' in model.datainfo.types and 'dv' in model.datainfo.types:
        maxobs = get_observations(model).max()
    else:
        maxobs = 1.0
    model = set_upper_bounds(model, {'POP_KM': 1.5 * maxobs})

    if denom != 1:
        if combined:
            cl = numer
            model, clmm = _add_parameter(model, 'CLMM', init=clmm_init)
        else:
            model = _rename_parameter(model, 'CL', 'CLMM')
            clmm = sympy.Symbol('CLMM')
            cl = 0
        vc = denom
    else:
        if combined:
            if model.statements.find_assignment('CL'):
                assignment = model.statements.find_assignment('CL')
                assert assignment is not None
                cl = assignment.symbol
            else:
                model, cl = _add_parameter(model, 'CL', clmm_init)
        else:
            cl = 0
        if model.statements.find_assignment('VC'):
            assignment = sset.find_assignment('VC')
            assert assignment is not None
            vc = assignment.symbol
        else:
            model, vc = _add_parameter(model, 'VC')  # FIXME: decide better initial estimate
        if not combined and model.statements.find_assignment('CL'):
            model = _rename_parameter(model, 'CL', 'CLMM')
            assignment = model.statements.find_assignment('CLMM')
            assert assignment is not None
            clmm = assignment.symbol
        else:
            model, clmm = _add_parameter(model, 'CLMM', init=clmm_init)

    rate = (clmm * km / (km + central.amount / vc) + cl) / vc
    cb = CompartmentalSystemBuilder(odes)
    cb.add_flow(central, output, rate)
    statements = (
        model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
    )
    statements = statements.remove_symbol_definitions(numer.free_symbols, statements.ode_system)
    model = model.replace(statements=statements)
    model = remove_unused_parameters_and_rvs(model)
    return model


def _rename_parameter(model: Model, old_name, new_name):
    statements = model.statements
    rvs = model.random_variables
    a = statements.find_assignment(old_name)
    assert a is not None
    d = {}
    for s in a.rhs_symbols:
        if s in model.parameters:
            old_par = s
            d[model.parameters[s].symbol] = f'POP_{new_name}'
            new_par = sympy.Symbol(f'POP_{new_name}')
            statements = statements.subs({old_par: new_par})
            break
    for s in a.rhs_symbols:
        iivs = rvs.iiv
        if s.name in iivs.names:
            cov = iivs.covariance_matrix
            ind = iivs.names.index(s.name)
            pars = [e for e in cov[ind, :] if e.is_Symbol]
            diag = cov[ind, ind]
            d[diag] = f'IIV_{new_name}'
            for p in pars:
                if p != diag:
                    if p.name.startswith('IIV'):
                        # FIXME: In some cases, parameters are read as symbols, this is
                        #  a workaround
                        try:
                            symb = p.symbol
                        except AttributeError:
                            symb = p
                        d[symb] = p.name.replace(f'IIV_{old_name}', f'IIV_{new_name}')
            rvs = rvs.subs(d)
            break
    new = []
    for p in model.parameters:
        if p.symbol in d:
            newparam = Parameter(
                name=d[p.symbol], init=p.init, upper=p.upper, lower=p.lower, fix=p.fix
            )
        else:
            newparam = p
        new.append(newparam)
    parameters = Parameters.create(new)
    statements = statements.subs({old_name: new_name})
    model = model.replace(statements=statements, parameters=parameters, random_variables=rvs)
    return model


def _get_mm_inits(model: Model, rate_numer, combined):
    pset, sset = model.parameters, model.statements
    parameter = _extract_params_from_symb(sset, rate_numer.name, pset)
    assert parameter is not None
    clmm_init = parameter.init

    if combined:
        clmm_init /= 2

    if model.dataset is not None and 'idv' in model.datainfo.types and 'dv' in model.datainfo.types:
        dv_max = get_observations(model).max()
    else:
        dv_max = 1.0
    km_init = dv_max / 2
    # FIXME: Cap initial estimate, this is NONMEM specific and should be handled more generally
    #  (https://github.com/pharmpy/pharmpy/issues/1395)
    if km_init >= 10**6:
        km_init = 5 * 10**5

    return km_init, clmm_init


def set_transit_compartments(model: Model, n: int, keep_depot: bool = True):
    """Set the number of transit compartments of model.

    Initial estimate for absorption rate is
    set the previous rate if available, otherwise it is set to the time of first observation/2.

    Parameters
    ----------
    model : Model
        Pharmpy model
    n : int
        Number of transit compartments
    keep_depot : bool
        False to convert depot compartment into a transit compartment

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_transit_compartments(model, 3)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → TRANSIT1
    ┌────────┐      ┌────────┐      ┌────────┐      ┌───────┐
    │TRANSIT1│──K12→│TRANSIT2│──K23→│TRANSIT3│──K34→│CENTRAL│──K40→
    └────────┘      └────────┘      └────────┘      └───────┘

    See also
    --------
    add_lag_time

    """
    statements = model.statements
    cs = statements.ode_system
    if cs is None:
        raise ValueError(f'Model {model.name} has no ODE system')
    transits = cs.find_transit_compartments(statements)
    try:
        n = _as_integer(n)
    except ValueError:
        raise ValueError(f'Number of compartments must be integer: {n}')

    model = remove_lag_time(model)

    # Handle keep_depot option
    depot = cs.find_depot(statements)
    mdt_init = None
    mdt_assign = None
    if not keep_depot and depot:
        central = cs.central_compartment
        rate = cs.get_flow(depot, central)
        assert rate is not None
        if not rate.is_Symbol:
            num, den = rate.as_numer_denom()
            if num == 1 and den.is_Symbol:
                symbol = den
            else:
                symbol = None
        else:
            symbol = rate
        if symbol:
            mdt_init = _extract_params_from_symb(statements, symbol.name, model.parameters).init
        inflows = cs.get_compartment_inflows(depot)
        cb = CompartmentalSystemBuilder(cs)
        if len(inflows) == 1:
            innode, inflow = inflows[0]
            cb.add_flow(innode, central, inflow)
        else:
            cb.set_dose(central, depot.doses[0])
        if statements.find_assignment('MAT'):
            model = _rename_parameter(model, 'MAT', 'MDT')
            statements = model.statements
            mdt_assign = statements.find_assignment('MDT')
        cb.remove_compartment(depot)
        statements = statements.before_odes + CompartmentalSystem(cb) + statements.after_odes
        statements = statements.remove_symbol_definitions(rate.free_symbols, statements.ode_system)
        if mdt_assign:
            statements = mdt_assign + statements
        model = model.replace(statements=statements)
        model = remove_unused_parameters_and_rvs(model)
        odes = statements.ode_system
        assert odes is not None
        # Since update_source() is used after removing the depot and statements are immutable, we need to
        # reset to get the correct rate names
        cs = model.statements.ode_system

    if len(transits) == n:
        return model
    elif n == 1 and has_instantaneous_absorption(model):
        raise ValueError(
            "Cannot set the number of transits to 1 for model with instantaneous "
            "absorption. The resulting model cannot be distinguished from first order absorption"
        )
    elif len(transits) == 0:
        if mdt_assign:
            mdt_symb = mdt_assign.symbol
        else:
            if mdt_init is not None:
                init = mdt_init
            else:
                init = _get_absorption_init(model, 'MDT')
            model, mdt_symb = _add_parameter(model, 'MDT', init=init)
        rate = n / mdt_symb
        dosing_comp = cs.dosing_compartments[0]
        comp = dosing_comp
        cb = CompartmentalSystemBuilder(cs)
        while n > 0:
            new_comp = Compartment.create(f'TRANSIT{n}')
            cb.add_compartment(new_comp)
            n -= 1
            cb.add_flow(new_comp, comp, rate)
            comp = new_comp
        comp = cb.set_bioavailability(comp, dosing_comp.bioavailability)
        dosing_comp = cb.set_bioavailability(dosing_comp, sympy.Integer(1))

        if len(dosing_comp.doses) == 1 and dosing_comp.doses[0].admid == 2:
            dosing_comp = cb.set_dose(dosing_comp, dosing_comp.doses[0].replace(admid=1))

        cb.move_dose(dosing_comp, comp, admid=1, replace=False)

        if len(dosing_comp.doses) == 1:
            cb.move_dose(dosing_comp, comp)
        else:
            dose = dosing_comp.sorted_doses(model)[0]
            cb.set_dose(comp, dose, replace=False)
            cb.set_dose(dosing_comp, dosing_comp.sorted_doses(model)[1:])

        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        model = model.replace(statements=statements)
        model = model.update_source()
    elif len(transits) > n:
        nremove = len(transits) - n
        removed_symbols = set()
        remaining = set(transits)
        trans, destination, flow = _find_last_transit(cs, remaining)

        cb = CompartmentalSystemBuilder(cs)

        while nremove > 0:
            try:
                from_comp, from_flow = cs.get_compartment_inflows(trans)[0]
            except IndexError:
                # Final remaining transit
                final_transit = True
            else:
                cb.add_flow(from_comp, destination, from_flow)
                final_transit = False
            cb.remove_compartment(trans)
            remaining.remove(trans)
            removed_symbols |= flow.free_symbols
            if not final_transit:
                trans = from_comp
                flow = from_flow
            nremove -= 1

        if n == 0:
            dose = cs.dosing_compartments[0].doses[0]
            cb.set_dose(destination, dose)

        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        model = model.replace(statements=statements)
        model = _update_numerators(model)
        statements = model.statements.remove_symbol_definitions(
            removed_symbols, model.statements.ode_system
        )
        model = model.replace(statements=statements)
        model = remove_unused_parameters_and_rvs(model)
    else:
        nadd = n - len(transits)
        last, destination, rate = _find_last_transit(cs, set(transits))
        cb = CompartmentalSystemBuilder(cs)
        cb.remove_flow(last, destination)
        while nadd > 0:
            new_comp = Compartment.create(f'TRANSIT{n - nadd + 1}')
            cb.add_compartment(new_comp)
            cb.add_flow(last, new_comp, rate)
            if rate.is_Symbol:
                ass = statements.find_assignment(rate.name)
                if ass is not None:
                    rate = ass.expression
            last = new_comp
            nadd -= 1
        cb.add_flow(last, destination, rate)
        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        model = model.replace(statements=statements)
        model = _update_numerators(model)
        model = model.update_source()
    return model


def _find_last_transit(odes: CompartmentalSystem, transits: Set[Compartment]):
    for trans in transits:
        destination, flow = odes.get_compartment_outflows(trans)[0]
        if destination not in transits:
            return trans, destination, flow

    raise ValueError('Could not find last transit')


def _update_numerators(model: Model):
    # update numerators for transit compartment rates
    statements = model.statements
    odes = statements.ode_system
    assert odes is not None
    transits = odes.find_transit_compartments(statements)
    new_numerator = sympy.Integer(len(transits))
    cb = CompartmentalSystemBuilder(odes)
    for comp in transits:
        to_comp, rate = odes.get_compartment_outflows(comp)[0]
        numer, denom = rate.as_numer_denom()
        if numer.is_Integer and numer != new_numerator:
            new_rate = new_numerator / denom
            cb.add_flow(comp, to_comp, new_rate)
        elif numer.is_Symbol:
            ass = statements.find_assignment(numer.name)
            if ass is not None:
                ass_numer, ass_denom = ass.expression.as_numer_denom()
                if ass_numer.is_Integer and ass_numer != new_numerator:
                    new_rate = new_numerator / ass_denom
                    statements = statements.reassign(numer, new_rate)
    model = model.replace(
        statements=statements.before_odes + CompartmentalSystem(cb) + statements.after_odes
    )
    return model


def add_lag_time(model: Model):
    """Add lag time to the dose compartment of model.

    Initial estimate for lag time is set the
    previous lag time if available, otherwise it is set to the time of first observation/2.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_lag_time(model)

    See also
    --------
    set_transit_compartments
    remove_lag_time

    """
    odes = get_and_check_odes(model)
    dosing_comp = odes.dosing_compartments[0]
    old_lag_time = dosing_comp.lag_time
    model, mdt_symb = _add_parameter(model, 'MDT', init=_get_absorption_init(model, 'MDT'))
    cb = CompartmentalSystemBuilder(odes)
    dosing_comp = cb.set_lag_time(dosing_comp, mdt_symb)

    # FIXME: Very temporary until new zo absorption logic is implemented
    if len(dosing_comp.doses) > 1:
        cb.set_lag_time(dosing_comp, sympy.Symbol("lag_time"))
        doses = dosing_comp.sorted_doses(model)
        oral_admid = doses[0].admid
        admid = sympy.Symbol("ADMID")
        model = model.replace(
            statements=(
                model.statements.before_odes
                + Assignment(
                    sympy.Symbol("lag_time"),
                    sympy.Piecewise((mdt_symb, sympy.Eq(admid, oral_admid)), (0, sympy.true)),
                )
                + CompartmentalSystem(cb)
                + model.statements.after_odes
            )
        )
    else:
        model = model.replace(
            statements=(
                model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
            )
        )
    if old_lag_time:
        model = model.replace(
            statements=model.statements.remove_symbol_definitions(old_lag_time.free_symbols, odes)
        )
        model = remove_unused_parameters_and_rvs(model)
    else:
        model = model.update_source()
    return model


def remove_lag_time(model: Model):
    """Remove lag time from the dose compartment of model.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = remove_lag_time(model)

    See also
    --------
    set_transit_compartments
    add_lag_time


    """
    odes = get_and_check_odes(model)
    dosing_comp = odes.dosing_compartments[0]
    lag_time = dosing_comp.lag_time
    if lag_time:
        symbols = lag_time.free_symbols
        cb = CompartmentalSystemBuilder(odes)
        cb.set_lag_time(dosing_comp, sympy.Integer(0))
        statements = (
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
        statements = statements.remove_symbol_definitions(symbols, statements.ode_system)
        model = model.replace(statements=statements)
        model = remove_unused_parameters_and_rvs(model)
    return model


def set_zero_order_absorption(model: Model):
    """Set or change to zero order absorption rate.

    Initial estimate for absorption rate is set
    the previous rate if available, otherwise it is set to the time of first observation/2.

    Parameters
    ----------
    model : Model
        Model to set or change to first order absorption rate

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_zero_order_absorption(model)
    >>> model.statements.ode_system
    Infusion(AMT, admid=1, duration=D1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CL/V→
    └───────┘

    See also
    --------
    set_instantaneous_absorption
    set_first_order_absorption

    """
    statements = model.statements
    odes = get_and_check_odes(model)
    _disallow_infusion(model, odes)
    assert isinstance(odes, CompartmentalSystem)

    if has_zero_order_absorption(model) and not has_seq_zo_fo_absorption(model):
        pass
    else:
        _disallow_infusion(model, odes)
        depot = odes.find_depot(statements)

        dose_comp = odes.dosing_compartments[0]
        symbols = dose_comp.free_symbols
        dose = dose_comp.sorted_doses(model)[0]
        lag_time = dose_comp.lag_time
        if depot:
            to_comp, _ = odes.get_compartment_outflows(depot)[0]
            ka = odes.get_flow(depot, odes.central_compartment)
            assert ka is not None
            cb = CompartmentalSystemBuilder(odes)
            cb.remove_compartment(depot)
            to_comp = cb.set_dose(to_comp, dose, replace=False)
            to_comp = cb.set_lag_time(to_comp, depot.lag_time)
            cb.set_bioavailability(to_comp, depot.bioavailability)
            statements = statements.before_odes + CompartmentalSystem(cb) + statements.after_odes
            symbols = ka.free_symbols
        new_statements = statements.remove_symbol_definitions(symbols, statements.ode_system)
        mat_idx = statements.find_assignment_index('MAT')
        if mat_idx is not None:
            # FIXME : Causes issue if mat_assign statement is dependent on previously
            # removed parameters/statements
            mat_assign = statements[mat_idx]
            new_statements = new_statements[0:mat_idx] + mat_assign + new_statements[mat_idx:]
        model = model.replace(statements=new_statements)

        model = remove_unused_parameters_and_rvs(model)
        if not has_zero_order_absorption(model):
            odes = model.statements.ode_system
            assert odes is not None
            model = _add_zero_order_absorption(
                model, dose, odes.dosing_compartments[0], 'MAT', lag_time
            )
            model = model.update_source()
        # FIXME : Very temporary until new zo absorption logic is implemented
        if lag_time != 0 and len(model.statements.ode_system.dosing_compartments[0].doses) > 1:
            model = remove_lag_time(model)
            model = add_lag_time(model)
    return model


def set_first_order_absorption(model: Model):
    """Set or change to first order absorption rate.

    Initial estimate for absorption rate is set to
    the previous rate if available, otherwise it is set to the time of first observation/2.

    If multiple doses is set to the affected compartment, currently only iv+oral
    doses (one of each) is supported

    Parameters
    ----------
    model : Model
        Model to set or change to use first order absorption rate

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_first_order_absorption(model)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → DEPOT
    ┌─────┐     ┌───────┐
    │DEPOT│──KA→│CENTRAL│──CL/V→
    └─────┘     └───────┘

    See also
    --------
    set_instantaneous_absorption
    set_zero_order_absorption

    """
    statements = model.statements
    cs = get_and_check_odes(model)

    if has_first_order_absorption(model) and not has_seq_zo_fo_absorption(model):
        pass
    else:
        depot = cs.find_depot(statements)
        dose_comp = cs.dosing_compartments[0]

        amount = dose_comp.doses[0].amount
        symbols = dose_comp.free_symbols
        lag_time = dose_comp.lag_time
        bio = dose_comp.bioavailability
        cb = CompartmentalSystemBuilder(cs)
        if depot and depot == dose_comp:
            dose_comp = cb.set_dose(
                dose_comp, Bolus(dose_comp.doses[0].amount), admid=dose_comp.doses[0].admid
            )
            dose_comp = cb.set_lag_time(dose_comp, sympy.Integer(0))
        if not depot:
            # TODO : Add another way of removing dependencies
            dose_admid = dose_comp.sorted_doses(model)[0].admid
            if len(dose_comp.doses) == 1:
                dose_comp = cb.set_dose(dose_comp, Bolus(amount))
                remove_dose = True
            else:
                dose_comp = cb.set_dose(dose_comp, dose_comp.sorted_doses(model)[1:])
                remove_dose = False

        statements = statements.before_odes + CompartmentalSystem(cb) + statements.after_odes
        new_statements = statements.remove_symbol_definitions(symbols, statements.ode_system)
        mat_idx = statements.find_assignment_index('MAT')
        if mat_idx is not None:
            mat_assign = statements[mat_idx]
            new_statements = new_statements[0:mat_idx] + mat_assign + new_statements[mat_idx:]

        model = model.replace(statements=new_statements)

        model = remove_unused_parameters_and_rvs(model)
        if not depot:
            # The new dose is created here
            model, _ = _add_first_order_absorption(
                model,
                Bolus(amount, admid=dose_admid),
                dose_comp,
                lag_time,
                bio,
                remove_dose=remove_dose,
            )
            model = model.update_source()
    return model


def set_instantaneous_absorption(model: Model):
    """Set or change to instantaneous absorption rate.

    Currently lagtime together with instantaneous absorption is not supported.

    Parameters
    ----------
    model : Model
        Model to set or change absorption rate

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_instantaneous_absorption(model)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CL/V→
    └───────┘

    See also
    --------
    set_zero_order_absorption
    set_first_order_absorption

    """
    statements = model.statements
    cs = get_and_check_odes(model)
    if has_instantaneous_absorption(model):
        pass
    else:
        depot = cs.find_depot(statements)
        if depot:
            to_comp, _ = cs.get_compartment_outflows(depot)[0]
            cb = CompartmentalSystemBuilder(cs)
            cb.set_dose(to_comp, depot.doses[0])
            ka = cs.get_flow(depot, cs.central_compartment)
            cb.remove_compartment(depot)
            symbols = ka.free_symbols
            statements = statements.before_odes + CompartmentalSystem(cb) + statements.after_odes
            model = model.replace(
                statements=statements.remove_symbol_definitions(symbols, statements.ode_system)
            )
            model = remove_unused_parameters_and_rvs(model)
        if has_zero_order_absorption(model):
            dose_comp = cs.dosing_compartments[0]
            old_symbols = dose_comp.free_symbols
            cb = CompartmentalSystemBuilder(cs)
            new_dose = Bolus(dose_comp.sorted_doses(model)[0].amount)
            if len(dose_comp.doses) > 1:
                cb.set_dose(dose_comp, (new_dose,) + dose_comp.sorted_doses(model)[1:])
            else:
                cb.set_dose(dose_comp, new_dose)
            unneeded_symbols = old_symbols - new_dose.free_symbols
            statements = statements.before_odes + CompartmentalSystem(cb) + statements.after_odes
            model = model.replace(
                statements=statements.remove_symbol_definitions(
                    unneeded_symbols, statements.ode_system
                )
            )
            model = remove_unused_parameters_and_rvs(model)
    return model


def set_seq_zo_fo_absorption(model: Model):
    """Set or change to sequential zero order first order absorption rate.

    Initial estimate for
    absorption rate is set the previous rate if available, otherwise it is set to the time of
    first observation/2.

    Currently lagtime together with sequential zero order first order absorption is not
    supported.

    Parameters
    ----------
    model : Model
        Model to set or change absorption rate

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_seq_zo_fo_absorption(model)
    >>> model.statements.ode_system
    Infusion(AMT, admid=1, duration=D1) → DEPOT
    ┌─────┐     ┌───────┐
    │DEPOT│──KA→│CENTRAL│──CL/V→
    └─────┘     └───────┘

    See also
    --------
    set_instantaneous_absorption
    set_zero_order_absorption
    set_first_order_absorption

    """
    statements = model.statements
    cs = get_and_check_odes(model)

    if has_seq_zo_fo_absorption(model):
        pass
    else:
        _disallow_infusion(model, cs)
        depot = cs.find_depot(statements)

        dose_comp = cs.dosing_compartments[0]
        have_ZO = has_zero_order_absorption(model)
        if depot and not have_ZO:
            model = _add_zero_order_absorption(model, dose_comp.doses[0], depot, 'MDT')
        elif not depot and have_ZO:
            if len(dose_comp.doses) == 1:
                fo_dose = dose_comp.doses[0]
                remove_dose = True
            else:
                fo_dose = dose_comp.sorted_doses(model)[0]
                cb = CompartmentalSystemBuilder(model.statements.ode_system)
                dose_comp = cb.set_dose(dose_comp, dose_comp.sorted_doses(model)[1:])
                model = model.replace(
                    statements=model.statements.before_odes
                    + CompartmentalSystem(cb)
                    + model.statements.after_odes
                )
                remove_dose = False
            model, _ = _add_first_order_absorption(
                model, fo_dose, dose_comp, remove_dose=remove_dose
            )
        elif not depot and not have_ZO:
            model = set_first_order_absorption(model)
            depot = model.statements.ode_system.find_depot(model.statements)
            model = _add_zero_order_absorption(
                model, Bolus(dose_comp.doses[0].amount), depot, 'MDT'
            )
        model = model.update_source()
    return model


def _disallow_infusion(model, odes):
    dose_comp = odes.dosing_compartments[0]
    if isinstance(dose_comp.doses[0], Infusion):
        if dose_comp.doses[0].rate is not None:
            ex = dose_comp.doses[0].rate
        else:
            ex = dose_comp.doses[0].duration
        assert ex is not None

        for s in ex.free_symbols:
            if s.name in model.datainfo.names:
                raise ModelError("Model already has an infusion given in the dataset")


def has_zero_order_absorption(model: Model):
    """Check if ode system describes a zero order absorption

    currently defined as having Infusion dose with rate not in dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_zero_order_absorption(model)
    False

    """
    cs = get_and_check_odes(model)
    dosing = cs.dosing_compartments[0]
    dose = dosing.doses[0]
    return _dose_zo(model, dose)


def _dose_zo(model, dose):
    if isinstance(dose, Infusion):
        if dose.rate is None:
            value = dose.duration
        else:
            value = dose.rate
        if isinstance(value, sympy.Symbol) or isinstance(value, str):
            name = str(value)
            if name not in model.datainfo.names:
                return True
        elif isinstance(value, sympy.Expr):
            assert value is not None
            names = {symb.name for symb in value.free_symbols}
            if not all(name in model.datainfo.names for name in names):
                return True
    return False


def has_first_order_absorption(model: Model):
    """Check if ode system describes a first order absorption

    Currently defined as the central compartment having a unidirectional input
    flow from another compartment (such as depot or transit)

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    -------
        Bool : True if model has first order absorption

    """

    odes = get_and_check_odes(model)
    dosing = odes.dosing_compartments[0]
    central = odes.central_compartment
    if dosing == central:
        return False

    in_flow = [flow[0] for flow in odes.get_compartment_inflows(central) if flow[0] != output]
    out_flow = [flow[0] for flow in odes.get_compartment_outflows(central) if flow[0] != output]

    unidirectional_flow = [uni_flow for uni_flow in in_flow if uni_flow not in out_flow]

    if len(unidirectional_flow) == 1:
        return True
    return False


def has_instantaneous_absorption(model: Model):
    """Check if ode system describes a instantaneous absorption

    Defined as being a instantaneous dose directly into the central compartment

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    -------
        Bool : True if model has instantaneous absorption
    """
    odes = get_and_check_odes(model)
    dosing = odes.dosing_compartments[0]
    central = odes.central_compartment
    if dosing != central:
        return False
    if isinstance(dosing.doses[0], Bolus):
        return True
    return False


def has_seq_zo_fo_absorption(model: Model):
    """Check if ode system describes a sequential zero-order, first-order absorption

    Defined as the model having both zero- and first-order absorption.

    Parameters
    ----------
    model : Model
        DPharmpy model

    See also
    --------
    has_zero_order_absorption
    has_first_order_absorption

    """
    if has_zero_order_absorption(model) and has_first_order_absorption(model):
        return True
    else:
        return False


def get_number_of_peripheral_compartments(model: Model):
    """Return the number of peripherals compartments connected to the central
    compartment

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    int
        Number of peripherals compartments
    """
    # Redundant function ?
    odes = get_and_check_odes(model)
    return len(odes.find_peripheral_compartments())


def get_number_of_transit_compartments(model: Model):
    """Return the number of transit compartments in the model

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    int
        Number of transit compartments
    """
    # Redundant function ?
    odes = get_and_check_odes(model)
    return len(odes.find_transit_compartments(model.statements))


def has_lag_time(model: Model):
    """Check if ode system is defined with absorption lag time

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    -------
        Bool : True if model is defined with lagtime
    """
    # Redundant function ?
    odes = model.statements.ode_system
    if odes is None:
        raise ValueError(f'Model {model.name} has no ODE system')

    dosing = odes.dosing_compartments[0]
    if dosing.lag_time:
        return True
    return False


def _add_zero_order_absorption(
    model, old_dose, to_comp, parameter_name, lag_time=None, replace=True
):
    """Add zero order absorption to a compartment. Initial estimate for absorption rate is set
    the previous rate if available, otherwise it is set to the time of first observation/2 is used.
    Disregards what is currently in the model.
    """
    mat_assign = model.statements.find_assignment(parameter_name)
    if mat_assign:
        mat_symb = mat_assign.symbol
    else:
        model, mat_symb = _add_parameter(
            model, parameter_name, init=_get_absorption_init(model, parameter_name)
        )
    new_dose = Infusion(old_dose.amount, admid=old_dose.admid, duration=mat_symb * 2)
    cb = CompartmentalSystemBuilder(model.statements.ode_system)
    dose_list = [new_dose] + list(to_comp.doses)
    dose_list.remove(old_dose)
    cb.set_dose(to_comp, tuple(dose_list), replace=replace)
    if lag_time is not None and lag_time != 0:
        cb.set_lag_time(model.statements.ode_system.dosing_compartments[0], lag_time)
    model = model.replace(
        statements=model.statements.before_odes
        + CompartmentalSystem(cb)
        + model.statements.after_odes
    )
    return model


def _add_first_order_absorption(
    model, dose, to_comp, lag_time=None, bioavailability=None, remove_dose=True
):
    """Add first order absorption
    Disregards what is currently in the model.
    """
    odes = model.statements.ode_system
    cb = CompartmentalSystemBuilder(odes)
    depot = Compartment.create(
        'DEPOT',
        doses=(dose,),
        lag_time=sympy.Integer(0) if lag_time is None else lag_time,
        bioavailability=sympy.Integer(1) if bioavailability is None else bioavailability,
    )
    cb.add_compartment(depot)
    if remove_dose:
        to_comp = cb.set_dose(to_comp, tuple())
    to_comp = cb.set_lag_time(to_comp, sympy.Integer(0))
    to_comp = cb.set_bioavailability(to_comp, sympy.Integer(1))

    mat_assign = model.statements.find_assignment('MAT')
    if mat_assign:
        mat_symb = mat_assign.symbol
    else:
        model, mat_symb = _add_parameter(model, 'MAT', _get_absorption_init(model, 'MAT'))
    cb.add_flow(depot, to_comp, 1 / mat_symb)
    model = model.replace(
        statements=model.statements.before_odes
        + CompartmentalSystem(cb)
        + model.statements.after_odes
    )
    return model, depot


def _get_absorption_init(model, param_name) -> float:
    try:
        if param_name == 'MDT':
            param_prev = model.statements.lag_time
        else:
            param_prev = _extract_params_from_symb(model.statements, param_name, model.parameters)
        return param_prev.init
    except (AttributeError, KeyError):
        pass

    try:
        time_label = model.datainfo.idv_column.name
    except IndexError:
        time_min = 1.0
    else:
        obs = get_observations(model)
        time = obs.index.get_level_values(level=time_label)
        time_min = time[time > 0].min()

    if param_name == 'MDT':
        init = float(time_min) / 2
    elif param_name == 'MAT':
        init = float(time_min) * 2
    else:
        raise NotImplementedError('param_name must be MDT or MAT')

    if init < 0:
        init = 0.1

    return init


def set_peripheral_compartments(model: Model, n: int, name: str = None):
    """Sets the number of peripheral compartments for central compartment to a specified number.

    If name is set, the peripheral compartment will be added to the compartment
    with the specified name instead.

    Parameters
    ----------
    model : Model
        Pharmpy model
    n : int
        Number of transit compartments
    name : str
        Name of compartment to add peripheral to.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_peripheral_compartments(model, 2)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────────┐
    │PERIPHERAL1│
    └───────────┘
      ↑      │
    Q2/V1  Q2/V2
      │      ↓
    ┌───────────┐
    │  CENTRAL  │──CL/V1→
    └───────────┘
       ↑      │
     Q3/V3  Q3/V1
       │      ↓
    ┌───────────┐
    │PERIPHERAL2│
    └───────────┘

    See also
    --------
    add_peripheral_compartment
    remove_peripheral_compartment

    """
    odes = model.statements.ode_system

    try:
        n = _as_integer(n)
    except TypeError:
        raise TypeError(f'Number of compartments must be integer: {n}')

    per = len(odes.find_peripheral_compartments())
    if per < n:
        for _ in range(n - per):
            model = add_peripheral_compartment(model, name=name)
    elif per > n:
        for _ in range(per - n):
            model = remove_peripheral_compartment(model, name=name)
    return model


def add_peripheral_compartment(model: Model, name: str = None):
    r"""Add a peripheral distribution compartment to model

    The rate of flow from the central to the peripheral compartment
    will be parameterized as QPn / VC where VC is the volume of the central compartment.
    The rate of flow from the peripheral to the central compartment
    will be parameterized as QPn / VPn where VPn is the volumne of the added peripheral
    compartment.

    If name is set, the peripheral compartment will be added to the compartment
    with the specified name instead.

    Initial estimates:

    ==  ===================================================
    n
    ==  ===================================================
    1   :math:`\mathsf{CL} = \mathsf{CL'}`, :math:`\mathsf{VC} = \mathsf{VC'}`,
        :math:`\mathsf{QP1} = \mathsf{CL'}` and :math:`\mathsf{VP1} = \mathsf{VC'} \cdot 0.05`
    2   :math:`\mathsf{QP1} = \mathsf{QP1' * 0.1}`, :math:`\mathsf{VP1} = \mathsf{VP1'}`,
        :math:`\mathsf{QP2} = \mathsf{QP1' * 0.9}` and :math:`\mathsf{VP2} = \mathsf{VP1'}`
    ==  ===================================================

    Parameters
    ----------
    model : Model
        Pharmpy model
    name : str
        Name of compartment to add peripheral to.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_peripheral_compartment(model)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────────┐
    │PERIPHERAL1│
    └───────────┘
      ↑      │
     Q/V1   Q/V2
      │      ↓
    ┌───────────┐
    │  CENTRAL  │──CL/V1→
    └───────────┘

    See also
    --------
    set_peripheral_compartment
    remove_peripheral_compartment

    """
    statements = model.statements
    odes = get_and_check_odes(model)
    if name:
        other_compartment = odes.find_compartment(name)
        if other_compartment is None:
            raise ValueError(f'{name} is not a compartment name.')
        per = odes.find_peripheral_compartments(name)
        central = other_compartment
    else:
        per = odes.find_peripheral_compartments()
        central = odes.central_compartment

    n = len(per) + 1

    elimination_rate = odes.get_flow(central, output)
    assert elimination_rate is not None
    if has_mixed_mm_fo_elimination(model):
        elimination_rate = sympy.expand(elimination_rate).args[0]
    cl, vc = elimination_rate.as_numer_denom()
    if cl.is_Symbol and vc == 1:
        # If K = CL / V
        s = statements.find_assignment(cl.name)
        assert s is not None
        cl, vc = s.expression.as_numer_denom()
    # Heuristic to handle the MM case
    if vc.is_Mul:
        vc = vc.args[0]

    cb = CompartmentalSystemBuilder(odes)

    # NOTE: Only used if vc != 1
    qp_init = 0.1
    vp_init = 0.1

    if n == 1:
        if vc != 1:
            # Heuristic to handle the Mixed MM-FO case
            if cl.is_Add:
                cl1 = cl.args[0]
                if cl1.is_Mul:
                    cl = cl1.args[0]
            pop_cl = _find_noncov_theta(model, cl)
            pop_vc = _find_noncov_theta(model, vc)
            pop_cl_init = model.parameters[pop_cl].init
            pop_vc_init = model.parameters[pop_vc].init
            qp_init = pop_cl_init
            vp_init = pop_vc_init * 0.05
    elif n == 2:
        if vc != 1:
            per1 = per[0]
            from_rate = odes.get_flow(per1, central)
            assert from_rate is not None
            qp1, vp1 = from_rate.as_numer_denom()
            if qp1.is_Symbol and vc == 1:
                # If K = CL / V
                s = statements.find_assignment(qp1.name)
                assert s is not None
                qp1, vp1 = s.expression.as_numer_denom()
            full_qp1 = statements.before_odes.full_expression(qp1)
            full_vp1 = statements.before_odes.full_expression(vp1)
            if full_vp1 == 1:
                full_qp1, full_vp1 = full_qp1.as_numer_denom()
            pop_qp1 = _find_noncov_theta(model, full_qp1, full=True)
            pop_vp1 = _find_noncov_theta(model, full_vp1, full=True)
            pop_qp1_init = model.parameters[pop_qp1].init
            pop_vp1_init = model.parameters[pop_vp1].init
            model = set_initial_estimates(model, {pop_qp1.name: pop_qp1_init * 0.10})
            qp_init = pop_qp1_init * 0.90
            vp_init = pop_vp1_init

    if vc != 1:
        model, qp = _add_parameter(model, f'QP{n}', init=qp_init)
        model, vp = _add_parameter(model, f'VP{n}', init=vp_init)
        if name:
            peripheral = Compartment.create(f'{name}_PERIPHERAL{n}')
        else:
            peripheral = Compartment.create(f'PERIPHERAL{n}')
        cb.add_compartment(peripheral)
        cb.add_flow(central, peripheral, qp / vc)
        cb.add_flow(peripheral, central, qp / vp)
    elif vc == 1:
        model, kpc = _add_parameter(model, f'KPC{n}', init=0.1)
        model, kcp = _add_parameter(model, f'KCP{n}', init=0.1)
        if name:
            peripheral = Compartment.create(f'{name}_PERIPHERAL{n}')
        else:
            peripheral = Compartment.create(f'PERIPHERAL{n}')
        cb.add_compartment(peripheral)
        cb.add_flow(central, peripheral, kcp)
        cb.add_flow(peripheral, central, kpc)

    model = model.replace(
        statements=Statements(
            model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
        )
    )

    return model.update_source()


def remove_peripheral_compartment(model: Model, name: str = None):
    r"""Remove a peripheral distribution compartment from model

    If name is set, a peripheral compartment will be removed from the compartment
    with the specified name.

    Initial estimates:

    ==  ===================================================
    n
    ==  ===================================================
    2   :math:`\mathsf{CL} = \mathsf{CL'}`,
        :math:`\mathsf{QP1} = \mathsf{CL'}` and :math:`\mathsf{VP1} = \mathsf{VC'} \cdot 0.05`
    3   :math:`\mathsf{QP1} = (\mathsf{QP1'} + \mathsf{QP2'}) / 2`,
        :math:`\mathsf{VP1} = \mathsf{VP1'} + \mathsf{VP2'}`
    ==  ===================================================

    Parameters
    ----------
    model : Model
        Pharmpy model
    name : str
        Name of compartment to remove peripheral compartment from.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_peripheral_compartments(model, 2)
    >>> model = remove_peripheral_compartment(model)
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────────┐
    │PERIPHERAL1│
    └───────────┘
      ↑       │
    Q/V1     Q/V2
      │       ↓
    ┌───────────┐
    │  CENTRAL  │──CL/V1→
    └───────────┘

    See also
    --------
    set_peripheral_compartment
    add_peripheral_compartment

    """
    odes = get_and_check_odes(model)
    peripherals = odes.find_peripheral_compartments(name)

    if peripherals:
        last_peripheral = peripherals[-1]
        if name:
            central = odes.find_compartment(name)
        else:
            central = odes.central_compartment

        if len(peripherals) == 1:
            # TODO: Elimination can be zero (drug metabolite)
            elimination_rate = odes.get_flow(central, output)
            if elimination_rate is sympy.Number(0):
                pass
            else:
                cl, vc = elimination_rate.as_numer_denom()
                if vc != 1:
                    from_rate = odes.get_flow(last_peripheral, central)
                    assert from_rate is not None

                    qp1, vp1 = from_rate.as_numer_denom()
                    pop_cl = _find_noncov_theta(model, cl)
                    pop_vc = _find_noncov_theta(model, vc)
                    pop_qp1 = _find_noncov_theta(model, qp1)
                    pop_vp1 = _find_noncov_theta(model, vp1)
                    pop_vc_init = model.parameters[pop_vc].init
                    pop_cl_init = model.parameters[pop_cl].init
                    pop_qp1_init = model.parameters[pop_qp1].init
                    pop_vp1_init = model.parameters[pop_vp1].init
                    new_vc_init = pop_vc_init + pop_qp1_init / pop_cl_init * pop_vp1_init
                    model = set_initial_estimates(model, {pop_vc.name: new_vc_init})
        elif len(peripherals) == 2:
            first_peripheral = peripherals[0]
            from1_rate = odes.get_flow(first_peripheral, central)
            assert from1_rate is not None
            qp1, vp1 = from1_rate.as_numer_denom()
            from2_rate = odes.get_flow(last_peripheral, central)
            assert from2_rate is not None
            qp2, vp2 = from2_rate.as_numer_denom()
            pop_qp2 = _find_noncov_theta(model, qp2)
            pop_vp2 = _find_noncov_theta(model, vp2)
            pop_qp1 = _find_noncov_theta(model, qp1)
            pop_vp1 = _find_noncov_theta(model, vp1)
            pop_qp2_init = model.parameters[pop_qp2].init
            pop_vp2_init = model.parameters[pop_vp2].init
            pop_qp1_init = model.parameters[pop_qp1].init
            pop_vp1_init = model.parameters[pop_vp1].init
            new_qp1_init = (pop_qp1_init + pop_qp2_init) / 2
            new_vp1_init = pop_vp1_init + pop_vp2_init
            model = set_initial_estimates(
                model, {pop_qp1.name: new_qp1_init, pop_vp1.name: new_vp1_init}
            )

        rate1 = odes.get_flow(central, last_peripheral)
        assert rate1 is not None
        rate2 = odes.get_flow(last_peripheral, central)
        assert rate2 is not None
        symbols = rate1.free_symbols | rate2.free_symbols
        cb = CompartmentalSystemBuilder(odes)
        cb.remove_compartment(last_peripheral)
        model = model.replace(
            statements=(
                model.statements.before_odes + CompartmentalSystem(cb) + model.statements.after_odes
            )
        )
        model = model.replace(
            statements=model.statements.remove_symbol_definitions(
                symbols, model.statements.ode_system
            )
        )
        model = remove_unused_parameters_and_rvs(model)
    return model


def set_ode_solver(
    model: Model, solver: Literal['CVODES', 'DGEAR', 'DVERK', 'IDA', 'LSODA', 'LSODI']
):
    """Sets ODE solver to use for model

    Recognized solvers and their corresponding NONMEM advans:

    +----------------------------+------------------+
    | Solver                     | NONMEM ADVAN     |
    +============================+==================+
    | CVODES                     | ADVAN14          |
    +----------------------------+------------------+
    | DGEAR                      | ADVAN8           |
    +----------------------------+------------------+
    | DVERK                      | ADVAN6           |
    +----------------------------+------------------+
    | IDA                        | ADVAN15          |
    +----------------------------+------------------+
    | LSODA                      | ADVAN13          |
    +----------------------------+------------------+
    | LSODI                      | ADVAN9           |
    +----------------------------+------------------+

    Parameters
    ----------
    model : Model
        Pharmpy model
    solver : {'CVODES', 'DGEAR', 'DVERK', 'IDA', 'LSODA', 'LSODI'}
        Solver to use or None for no preference

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_ode_solver(model, 'LSODA')

    """
    new_steps = []
    for step in model.estimation_steps:
        new = step.replace(solver=solver)
        new_steps.append(new)
    newsteps = EstimationSteps.create(new_steps)
    model = model.replace(estimation_steps=newsteps).update_source()
    return model


def find_clearance_parameters(model: Model):
    """Find clearance parameters in model

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    list
        A list of clearance parameters

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> find_clearance_parameters(model)
    [CL]
    """
    cls = set()
    sset = model.statements
    odes = get_and_check_odes(model)
    t = odes.t
    rate_list = _find_rate(model, sset)
    for rate in rate_list:
        if isinstance(rate, sympy.Symbol):
            assignment = sset.find_assignment(rate)
            assert assignment is not None
            rate = assignment.expression
        a, b = map(lambda x: x.free_symbols, rate.as_numer_denom())
        if b:
            clearance_symbols = a - b - {t}
            for clearance in clearance_symbols:
                clearance = _find_real_symbol(sset, clearance)
                if str(clearance) not in ['LAFREE', 'KINT', 'KON']:  # exclude TMDD parameters
                    cls.add(clearance)
    return sorted(cls, key=str)


def find_volume_parameters(model: Model):
    """Find volume parameters in model

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    list
        A list of volume parameters

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> find_volume_parameters(model)
    [V]
    """
    vcs = set()
    sset = model.statements
    odes = sset.ode_system
    if odes is None:
        raise ValueError(f'Model {model.name} has no ODE system')
    t = odes.t
    rate_list = _find_rate(model, sset)
    for rate in rate_list:
        if isinstance(rate, sympy.Symbol):
            assignment = sset.find_assignment(rate)
            assert assignment is not None
            rate = assignment.expression
        rate = sympy.cancel(rate)
        a, b = map(lambda x: x.free_symbols, rate.as_numer_denom())
        volume_symbols = b - a - {t}
        for volume in volume_symbols:
            volume = _find_real_symbol(sset, volume)
            vcs.add(volume)
    return sorted(vcs, key=str)


def _find_rate(model: Model, sset: Statements):
    rate_list = []
    odes = sset.ode_system
    assert isinstance(odes, CompartmentalSystem)
    central = odes.central_compartment
    # FIXME : How to specify metabolite compartment?
    metacomp = odes.find_compartment("METABOLITE")
    if metacomp:
        central_to_metabolite = odes.get_flow(central, metacomp)
        rate_list.append(central_to_metabolite)
        elimination_rate = odes.get_flow(metacomp, output)
        rate_list.append(elimination_rate)

    else:
        elimination_rate = odes.get_flow(central, output)
        rate_list.append(elimination_rate)
        if has_mixed_mm_fo_elimination(model):
            elimination_rate = sympy.expand(elimination_rate)
            rate_list.append(elimination_rate.args[0])
            rate_list.append(elimination_rate.args[1])

    comp_peripherals = {}
    comp_peripherals[central] = odes.find_peripheral_compartments()
    if metacomp:
        comp_peripherals[metacomp] = odes.find_peripheral_compartments(name=metacomp.name)

    for comp, peripherals in comp_peripherals.items():
        for periph in peripherals:
            rate1 = odes.get_flow(comp, periph)
            rate_list.append(rate1)
            rate2 = odes.get_flow(periph, comp)
            rate_list.append(rate2)
    return rate_list


def _find_real_symbol(sset, symbol):
    assign = sset.find_assignment(symbol)
    if len(assign.rhs_symbols) == 1:
        dep_assigns = _get_dependent_assignments(sset, assign)
        for dep_assign in dep_assigns:
            symbol = dep_assign.symbol
    return symbol


def _get_dependent_assignments(sset, assignment):
    """Finds dependent assignments one layer deep"""
    return list(
        filter(
            None,  # NOTE: filters out falsy values
            (sset.find_assignment(symb) for symb in assignment.expression.free_symbols),
        )
    )


def has_odes(model: Model) -> bool:
    """Check if model has an ODE system

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        True if model has an ODE system

    See also
    --------
    has_linear_odes
        has_linear_odes_with_real_eigenvalues

    Examples
    --------
    >>> from pharmpy.modeling import has_odes, load_example_model
    >>> model = load_example_model("pheno")
    >>> has_odes(model)
    True
    """

    odes = model.statements.ode_system
    return bool(odes)


def has_linear_odes(model: Model) -> bool:
    """Check if model has a linear ODE system

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        True if model has an ODE system that is linear

    See also
    --------
    has_odes
        has_linear_odes_with_real_eigenvalues

    Examples
    --------
    >>> from pharmpy.modeling import has_linear_odes, load_example_model
    >>> model = load_example_model("pheno")
    >>> has_linear_odes(model)
    True
    """

    if not has_odes(model):
        return False

    odes = model.statements.ode_system
    assert isinstance(odes, CompartmentalSystem)
    symbs = odes.compartmental_matrix.free_symbols | odes.zero_order_inputs.free_symbols
    return odes.t not in symbs


def has_linear_odes_with_real_eigenvalues(model: Model):
    """Check if model has a linear ode system with real eigenvalues

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        True if model has an ODE system that is linear

    See also
    --------
    has_odes
        has_linear_odes

    Examples
    --------
    >>> from pharmpy.modeling import has_linear_odes_with_real_eigenvalues, load_example_model
    >>> model = load_example_model("pheno")
    >>> has_linear_odes_with_real_eigenvalues(model)
    True
    """

    if not has_odes(model):
        return False
    if not has_linear_odes(model):
        return False
    odes = model.statements.ode_system
    M = odes.compartmental_matrix
    eigs = M.eigenvals().keys()
    for eig in eigs:
        real = is_real(model, eig)
        if real is None or not real:
            return real
    return True


def get_initial_conditions(
    model: Model, dosing: bool = False
) -> Mapping[sympy.Function, sympy.Expr]:
    """Get initial conditions for the ode system

    Default initial conditions at t=0 for amounts is 0

    Parameters
    ----------
    model : Model
        Pharmpy model
    dosing : bool
        Set to True to add dosing as initial conditions

    Return
    ------
    dict
        Initial conditions

    Examples
    --------
    >>> from pharmpy.modeling import get_initial_conditions, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_initial_conditions(model)
    {A_CENTRAL(0): 0}
    >>> get_initial_conditions(model, dosing=True)
    {A_CENTRAL(0): AMT}
    """
    d = {}
    odes = model.statements.ode_system
    if odes is None:
        return d
    assert isinstance(odes, CompartmentalSystem)
    for amt in odes.amounts:
        d[sympy.Function(amt.name)(0)] = sympy.Integer(0)
    for s in model.statements:
        if isinstance(s, Assignment):
            if s.symbol.is_Function and not (s.symbol.args[0].free_symbols):
                d[s.symbol] = s.expression

    if dosing:
        for name in odes.compartment_names:
            comp = odes.find_compartment(name)
            if comp.doses and isinstance(comp.doses[0], Bolus):
                if comp.lag_time:
                    time = comp.lag_time
                else:
                    time = 0
                d[sympy.Function(comp.amount.name)(time)] = comp.doses[0].amount

    return d


def set_initial_condition(
    model: Model,
    compartment: str,
    expression: Union[str, int, float, sympy.Expr],
    time: Union[str, int, float, sympy.Expr] = sympy.Integer(0),
) -> Model:
    """Set an initial condition for the ode system

    If the initial condition is already set it will be updated. If the initial condition
    is set to zero at time zero it will be removed (since the default is 0).

    Parameters
    ----------
    model : Model
        Pharmpy model
    compartment : str
        Name of the compartment
    expression : Union[str, sympy.Expr]
        The expression of the initial condition
    time : Union[str, sympy.Expr]
        Time point. Default 0

    Return
    ------
    model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_initial_condition(model, "CENTRAL", 10)
    >>> get_initial_conditions(model)
    {A_CENTRAL(0): 10}
    """
    odes = model.statements.ode_system
    if odes is None:
        raise ValueError("Model has no system of ODEs")
    comp = odes.find_compartment(compartment)
    if comp is None:
        raise ValueError(f"Model has no compartment named {compartment}")
    expr = sympy.sympify(expression)
    time = sympy.sympify(time)
    amount = sympy.Function(comp.amount.name)(time)
    assignment = Assignment(amount, expr)
    for i, s in enumerate(model.statements.before_odes):
        if s.symbol == amount:
            if time == 0 and expr == 0:
                statements = model.statements[:i] + model.statements[i + 1 :]
            else:
                statements = model.statements[:i] + assignment + model.statements[i + 1 :]
            break
    else:
        if not (time == 0 and expr == 0):
            statements = (
                model.statements.before_odes + assignment + odes + model.statements.after_odes
            )
    model = model.replace(statements=statements)
    return model.update_source()


def get_zero_order_inputs(model: Model) -> sympy.Matrix:
    """Get zero order inputs for all compartments

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    sympy.Matrix
        Vector of inputs

    Examples
    --------
    >>> from pharmpy.modeling import get_zero_order_inputs, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_zero_order_inputs(model)
    Matrix([[0]])
    """
    odes = model.statements.ode_system
    if odes is None:
        return sympy.Matrix()
    return odes.zero_order_inputs


def set_zero_order_input(
    model: Model, compartment: str, expression: Union[str, int, float, sympy.Expr]
) -> Model:
    """Set a zero order input for the ode system

    If the zero order input is already set it will be updated.

    Parameters
    ----------
    model : Model
        Pharmpy model
    compartment : str
        Name of the compartment
    expression : Union[str, sympy.Expr]
        The expression of the zero order input

    Return
    ------
    model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_zero_order_input(model, "CENTRAL", 10)
    >>> get_zero_order_inputs(model)
    Matrix([[10]])
    """
    odes = model.statements.ode_system
    if odes is None:
        raise ValueError("Model has no system of ODEs")
    comp = odes.find_compartment(compartment)
    if comp is None:
        raise ValueError(f"Model has no compartment named {compartment}")
    expr = sympy.sympify(expression)
    cb = CompartmentalSystemBuilder(odes)
    cb.set_input(comp, expr)
    cs = CompartmentalSystem(cb)
    statements = model.statements.before_odes + cs + model.statements.after_odes
    model = model.replace(statements=statements)
    return model.update_source()


class ODEDisplayer:
    def __init__(self, eqs, ics):
        self._eqs = eqs
        self._ics = ics

    def __repr__(self):
        if self._eqs is None:
            return ""
        a = []
        for ode in self._eqs:
            ode_str = sympy.pretty(ode)
            a += ode_str.split('\n')
        for key, value in self._ics.items():
            ics_str = sympy.pretty(sympy.Eq(key, value))
            a += ics_str.split('\n')
        return bracket(a)

    def _repr_latex_(self):
        if self._eqs is None:
            return ""
        rows = []
        for ode in self._eqs:
            ode_repr = sympy.latex(ode, mul_symbol='dot')
            rows.append(ode_repr)
        for k, v in self._ics.items():
            ics_eq = sympy.Eq(k, v)
            ics_repr = sympy.latex(ics_eq, mul_symbol='dot')
            rows.append(ics_repr)
        return r'\begin{cases} ' + r' \\ '.join(rows) + r' \end{cases}'


def display_odes(model: Model):
    """Displays the ordinary differential equation system

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    ODEDisplayer
        A displayable object

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> display_odes(model)
    ⎧d                  -CL⋅A_CENTRAL(t)
    ⎨──(A_CENTRAL(t)) = ─────────────────
    ⎩dt                         V
    <BLANKLINE>

    """

    odes = model.statements.ode_system
    if odes is not None:
        eqs = odes.eqs
        ics = {key: val for key, val in get_initial_conditions(model).items() if val != 0}
    else:
        eqs = None
        ics = None
    return ODEDisplayer(eqs, ics)


def solve_ode_system(model: Model):
    """Replace ODE system with analytical solution if possible

    Warnings
    --------
    This function can currently only handle the most simple of ODE systems.

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Model
        Pharmpy model object

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model.statements.ode_system
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CL/V→
    └───────┘
    >>> model = solve_ode_system(model)

    """
    odes = model.statements.ode_system
    if odes is None:
        return model
    ics = get_initial_conditions(model, dosing=True)
    # FIXME: Should set assumptions on symbols before solving
    # FIXME: Need a way to handle systems with no explicit solutions
    sol = sympy.dsolve(odes.eqs, ics=ics)
    new = []
    for s in model.statements:
        if isinstance(s, CompartmentalSystem):
            for eq in sol:
                ass = Assignment(eq.lhs, eq.rhs)
                new.append(ass)
        else:
            new.append(s)
    model = model.replace(statements=Statements(new)).update_source()
    return model


def get_central_volume_and_clearance(model: Model):
    """Get the volume and clearance parameters

    Parameters
    ----------
    model: Model
        Pharmpy model


    Returns
    -------
    Sympy.Symbol
        Volume symbol
    Sympy.Symbol
        Clearance symbol

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> get_central_volume_and_clearance(model)
    (V, CL)

    """
    vcs = set()
    cls = set()
    sset = model.statements
    odes = sset.ode_system
    if odes is None:
        raise ValueError(f'Model {model.name} has no ODE system')
    t = odes.t
    odes = model.statements.ode_system
    central_comp = odes.central_compartment
    rate = odes.get_flow(central_comp, output)
    rate = sympy.expand(rate)
    if has_mixed_mm_fo_elimination(model):
        rate = rate.args[0]
    if isinstance(rate, sympy.Symbol):
        assignment = sset.find_assignment(rate)
        assert assignment is not None
        rate = assignment.expression
    rate = sympy.cancel(rate)
    a, b = map(lambda x: x.free_symbols, rate.as_numer_denom())
    if b:
        # Get volume parameter
        volume_symbols = b - a - {t}
        for volume in volume_symbols:
            volume = _find_real_symbol(sset, volume)
            vcs.add(volume)
        # Get clearance parameter
        clearance_symbols = a - b - {t}
        for clearance in clearance_symbols:
            clearance = _find_real_symbol(sset, clearance)
            cls.add(clearance)
    else:
        raise ValueError('Model is not suitable')
    return list(vcs)[0], list(cls)[0]



from pharmpy.model import Parameters
from pharmpy.modeling import load_example_model
from pharmpy.tools.rankfuncs import aic, ofv

pheno = load_example_model("pheno")


class DummyResults:
    def __init__(self, ofv=None, parameter_estimates=None):
        self.ofv = ofv
        self.parameter_estimates = parameter_estimates

    def __bool__(self):
        return bool(self.ofv) and bool(self.parameter_estimates)


class DummyModel:
    def __init__(self, name, parameters=None, **kwargs):
        self.name = name
        self.parameters = parameters
        self.dataset = pheno.dataset
        self.datainfo = pheno.datainfo
        if 'no_modelfit_results' in kwargs.keys():
            self.modelfit_results = None
        else:
            self.modelfit_results = DummyResults(**kwargs)


def test_ofv():
    run1 = DummyModel("run1", ofv=0)
    run2 = DummyModel("run2", ofv=-1)
    run3 = DummyModel("run3", ofv=-14)
    res, _ = ofv(run1, [run2, run3])
    assert [run3] == res

    run4 = DummyModel("run4", ofv=2)
    run5 = DummyModel("run5", ofv=-2)
    res, _ = ofv(run1, [run2, run3, run4, run5], cutoff=2)
    assert [run3, run5] == res

    run6 = DummyModel("run6", no_modelfit_results=True)
    res, _ = ofv(run6, [run1, run2, run3])
    assert [] == res

    run7 = DummyModel("run7", no_modelfit_results=True)
    run8 = DummyModel("run8", no_modelfit_results=True)
    res, _ = ofv(run6, [run7, run8])
    assert [] == res


def test_aic():
    run1 = DummyModel("run1", ofv=0, parameters=Parameters())
    run2 = DummyModel("run2", ofv=-1, parameters=Parameters())
    run3 = DummyModel("run3", ofv=-14, parameters=Parameters())
    res, _ = aic(run1, [run2, run3], cutoff=3.84)
    assert [run3] == res
    res, _ = aic(run1, [run2, run3])
    assert [run3, run2] == res



class RawRecord:
    """A record that just keeps contents unparsed.
    Used for unknown records and for anything coming before the first record
    """

    def __init__(self, content, name='', raw_name=''):
        self.name = name
        self.raw_name = raw_name
        self.content = content

    def __str__(self):
        return self.raw_name + self.content



import pharmpy.model
from pharmpy.deps import sympy
from pharmpy.model import Assignment, EstimationStep, EstimationSteps, Statements
from pharmpy.workflows import Task, Workflow, WorkflowBuilder


def create_workflow(model=None):
    wb = WorkflowBuilder(name="linearize")

    if model is not None:
        start_task = Task('start_linearize', start_linearize, model)
    else:
        start_task = Task('start_linearize', start_linearize)

    wb.add_task(start_task)
    return Workflow(wb)


def start_linearize(model):
    return model


def create_linearized_model(model):
    linbase = pharmpy.model.Model(
        parameters=model.parameters,
        random_variables=model.random_variables,
        datainfo=model.datainfo,
    )

    ms = []
    base_terms_sum = 0
    for i, eta in enumerate(model.random_variables.etas.names, start=1):
        deta = sympy.Symbol("D_ETA1")
        oeta = sympy.Symbol("OETA")
        base = Assignment(sympy.Symbol(f'BASE{i}'), deta * (sympy.Symbol(eta) - oeta))
        ms.append(base)
        base_terms_sum += base.symbol

    base_terms = Assignment(sympy.Symbol('BASE_TERMS'), base_terms_sum)
    ms.append(base_terms)
    ipred = Assignment(sympy.Symbol('IPRED'), sympy.Symbol('OPRED') + base_terms.symbol)
    ms.append(ipred)

    i = 1
    err_terms_sum = 0
    for epsno, eps in enumerate(model.random_variables.epsilons, start=1):
        err = Assignment(sympy.Symbol(f'ERR{epsno}'), sympy.Symbol(f'D_EPS{epsno}'))
        err_terms_sum += err.symbol
        ms.append(err)
        i += 1
        for etano, eta in enumerate(model.random_variables.etas.names, start=1):
            inter = Assignment(
                sympy.Symbol(f'ERR{i}'),
                sympy.Symbol(f'D_EPSETA{epsno}_{etano}')
                * (sympy.Symbol(eta) - sympy.Symbol(f'OETA{etano}')),
            )
            err_terms_sum += inter.symbol
            ms.append(inter)
            i += 1
    error_terms = Assignment(sympy.Symbol('ERROR_TERMS'), err_terms_sum)
    ms.append(error_terms)

    # FIXME: Handle other DVs?
    y = list(model.dependent_variables.keys())[0]
    Assignment(y, ipred.symbol + error_terms.symbol)

    est = EstimationStep.create('foce', interaction=True)
    linbase = linbase.replace(
        name='linbase', statements=Statements(ms), estimation_steps=EstimationSteps.create([est])
    )
    return linbase



import itertools

from pharmpy.modeling import add_estimation_step, remove_estimation_step, set_ode_solver
from pharmpy.tools.common import update_initial_estimates
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder


def exhaustive(methods, solvers, parameter_uncertainty_methods):
    wb = WorkflowBuilder()

    task_start = Task('start', start)
    wb.add_task(task_start)

    add_eval_after_est = len(methods) > 1

    candidate_no = 1
    for method, solver, parameter_uncertainty_method in itertools.product(
        methods, solvers, parameter_uncertainty_methods
    ):
        model_name = f'estmethod_run{candidate_no}'
        task_create_candidate = Task(
            'create_candidate',
            _create_candidate_model,
            model_name,
            method,
            solver,
            parameter_uncertainty_method,
            add_eval_after_est,
            False,
            False,
        )

        wb.add_task(task_create_candidate, predecessors=task_start)
        candidate_no += 1

    return Workflow(wb), None


def exhaustive_with_update(methods, solvers, parameter_uncertainty_methods):
    wb = WorkflowBuilder()

    add_eval_after_est = len(methods) > 1

    parameter_uncertainty_method_base = (
        parameter_uncertainty_methods[0] if 'FOCE' in methods else None
    )

    task_base_model = Task(
        'create_base_model',
        _create_base_model,
        parameter_uncertainty_method_base,
        add_eval_after_est,
    )
    wb.add_task(task_base_model)
    wf_fit = create_fit_workflow(n=1)
    wb.insert_workflow(wf_fit, predecessors=task_base_model)
    task_base_model_fit = wb.output_tasks

    candidate_no = 1
    for method, solver, parameter_uncertainty_method in itertools.product(
        methods, solvers, parameter_uncertainty_methods
    ):
        # This is equivalent to the base model
        if not (
            method == 'FOCE'
            and solver is None
            and parameter_uncertainty_method == parameter_uncertainty_methods[0]
        ):
            model_name = f'estmethod_run{candidate_no}'
            task_create_candidate = Task(
                'create_candidate',
                _create_candidate_model,
                model_name,
                method,
                solver,
                parameter_uncertainty_method,
                add_eval_after_est,
                False,
                False,
            )
            wb.add_task(task_create_candidate, predecessors=task_base_model_fit)
            candidate_no += 1

        # Create model with updated estimates from FOCE
        model_name = f'estmethod_run{candidate_no}'
        task_create_candidate = Task(
            'create_candidate',
            _create_candidate_model,
            model_name,
            method,
            solver,
            parameter_uncertainty_method,
            add_eval_after_est,
            True,
            False,
        )
        wb.add_task(task_create_candidate, predecessors=task_base_model_fit)
        candidate_no += 1

    return Workflow(wb), task_base_model_fit


def exhaustive_only_eval(methods, solvers, parameter_uncertainty_methods):
    wb = WorkflowBuilder()

    task_start = Task('start', start)
    wb.add_task(task_start)

    add_eval_after_est = len(methods) > 1

    candidate_no = 1
    for method, solver, parameter_uncertainty_method in itertools.product(
        methods, solvers, parameter_uncertainty_methods
    ):
        model_name = f'estmethod_run{candidate_no}'
        task_create_candidate = Task(
            'create_candidate',
            _create_candidate_model,
            model_name,
            method,
            solver,
            parameter_uncertainty_method,
            add_eval_after_est,
            False,
            True,
        )

        wb.add_task(task_create_candidate, predecessors=task_start)
        candidate_no += 1

    return Workflow(wb), None


def start(model_entry):
    return model_entry


def _create_base_model(parameter_uncertainty_method, add_eval_after_est, model_entry):
    model = model_entry.model

    est_settings = _create_est_settings("FOCE", parameter_uncertainty_method, add_eval_after_est)
    est_method = est_settings["method"]
    all_methods = [est_method]

    if add_eval_after_est:
        eval_settings = _create_eval_settings(
            laplace=False, parameter_uncertainty_method=parameter_uncertainty_method
        )
        eval_method = eval_settings["method"]
        all_methods.append(eval_method)
    else:
        eval_settings = eval_method = None

    base_model = model.replace(name="base_model")

    base_model = base_model.replace(
        description=_create_description(
            methods=all_methods,
            parameter_uncertainty_method=parameter_uncertainty_method,
            solver=None,
            update_inits=False,
        )
    )

    while len(base_model.estimation_steps) > 0:
        base_model = remove_estimation_step(base_model, 0)

    base_model = add_estimation_step(base_model, **est_settings)
    if add_eval_after_est:
        base_model = add_estimation_step(base_model, **eval_settings)

    base_entry = ModelEntry.create(base_model, modelfit_results=None, parent=model_entry.model)
    return base_entry


def _create_candidate_model(
    model_name,
    method,
    solver,
    parameter_uncertainty_method,
    add_eval_after_est,
    update_inits,
    only_evaluation,
    model_entry,
):
    if update_inits:
        model = update_initial_estimates(model_entry.model, model_entry.modelfit_results)
    else:
        model = model_entry.model

    est_settings = None
    eval_settings = None

    laplace = True if method == 'LAPLACE' else False

    if only_evaluation:
        eval_settings = _create_eval_settings(laplace, parameter_uncertainty_method)
        all_methods = [eval_settings['method']]
    else:
        est_settings = _create_est_settings(
            method, parameter_uncertainty_method, add_eval_after_est
        )
        all_methods = [est_settings['method']]

        if add_eval_after_est:
            eval_settings = _create_eval_settings(laplace, parameter_uncertainty_method)
            all_methods.append(eval_settings['method'])

    model = model.replace(
        name=model_name,
        description=_create_description(
            methods=all_methods,
            solver=solver,
            parameter_uncertainty_method=parameter_uncertainty_method,
            update_inits=update_inits,
        ),
    )

    while len(model.estimation_steps) > 0:
        model = remove_estimation_step(model, 0)

    if est_settings:
        model = add_estimation_step(model, **est_settings)
    if eval_settings:
        model = add_estimation_step(model, **eval_settings)

    if solver:
        model = set_ode_solver(model, solver)

    candidate_entry = ModelEntry.create(model, modelfit_results=None, parent=model_entry.model)
    return candidate_entry


def _create_est_settings(method, parameter_uncertainty_method=None, add_eval_after_est=False):
    est_settings = {
        'method': method,
        'interaction': True,
        'laplace': False,
        'auto': True,
        'keep_every_nth_iter': 10,
        'maximum_evaluations': 9999,
    }

    if method == 'LAPLACE':
        est_settings['method'] = 'FOCE'
        est_settings['laplace'] = True

    if add_eval_after_est:
        est_settings['parameter_uncertainty_method'] = None
    else:
        est_settings['parameter_uncertainty_method'] = parameter_uncertainty_method

    return est_settings


def _create_eval_settings(laplace=False, parameter_uncertainty_method=None):
    eval_settings = {
        'method': 'IMP',
        'interaction': True,
        'evaluation': True,
        'laplace': False,
        'maximum_evaluations': 9999,
        'isample': 10000,
        'niter': 10,
        'keep_every_nth_iter': 10,
        'parameter_uncertainty_method': parameter_uncertainty_method,
    }

    if laplace:
        eval_settings['laplace'] = True

    return eval_settings


def _create_description(methods, solver, parameter_uncertainty_method, update_inits=False):
    model_description = ','.join(methods)
    if solver:
        model_description += f';{solver}'
    if parameter_uncertainty_method:
        model_description += f';{parameter_uncertainty_method}'
    if update_inits:
        model_description += ' (update_inits)'
    return model_description



from dataclasses import dataclass
from typing import Literal, Tuple, Union

from .count_interpreter import CountInterpreter
from .feature import ModelFeature, feature
from .symbols import Name, Wildcard


@dataclass(frozen=True)
class Transits(ModelFeature):
    counts: Tuple[int, ...]
    depot: Union[Tuple[Name[Literal['DEPOT', 'NODEPOT']], ...], Wildcard] = (Name('DEPOT'),)

    def __add__(self, other):
        # NOTE : Use with caution as no addition logic is implemented here
        if isinstance(self.depot, Wildcard) or isinstance(other.depot, Wildcard):
            depot = Wildcard()
        else:
            depot = tuple(set(self.depot + tuple([a for a in other.depot if a not in self.depot])))

        counts = tuple(set(self.counts + tuple([a for a in other.counts if a not in self.counts])))

        return Transits(counts=counts, depot=depot)

    def __sub__(self, other):
        # NOTE : Use with caution as no subtraction logic is implemented here
        if isinstance(other.depot, Wildcard):
            all_depot = (Name('DEPOT'),)
        elif isinstance(self.depot, Wildcard):
            default = self._wildcard
            all_depot = tuple([a for a in default if a not in other.depot])
        else:
            all_depot = tuple([a for a in self.depot if a not in other.depot])

        if len(all_depot) == 0:
            all_depot = (Name('DEPOT'),)

        all_counts = tuple([a for a in self.counts if a not in other.counts])

        if len(all_counts) == 0:
            all_counts = (0,)

        return Transits(counts=all_counts, depot=all_depot)

    def __eq__(self, other):
        if isinstance(other, Transits):
            if isinstance(self.depot, Wildcard):
                lhs_depot = self.depot
            else:
                lhs_depot = set(self.depot)

            if isinstance(other.depot, Wildcard):
                rhs_depot = other.depot
            else:
                rhs_depot = set(other.depot)

            return (set(self.counts) == set(other.counts), lhs_depot == rhs_depot)
        else:
            return False

    @property
    def eval(self):
        if isinstance(self.depot, Wildcard):
            return Transits(self.counts, self._wildcard)
        else:
            return self

    @property
    def _wildcard(self):
        return tuple([Name(x) for x in ['DEPOT', 'NODEPOT']])


class TransitsInterpreter(CountInterpreter):
    def interpret(self, tree):
        children = self.visit_children(tree)
        assert 1 <= len(children) <= 2
        return feature(Transits, children)

    def depot_modes(self, tree):
        children = self.visit_children(tree)
        return list(Name(child.value.upper()) for child in children)

    def depot_wildcard(self, tree):
        return Wildcard()



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy


def free_images(expr: sympy.Expr) -> set[sympy.Expr]:
    return set(_free_images_iter(expr))


def free_images_and_symbols(expr: sympy.Expr) -> set[sympy.Expr]:
    return expr.free_symbols | free_images(expr)  # pyright: ignore [reportGeneralTypeIssues]


def _free_images_iter(expr: sympy.Expr) -> Iterable[sympy.Expr]:
    if isinstance(expr.func, sympy.core.function.UndefinedFunction):
        yield expr
        return

    yield from chain.from_iterable(
        map(
            _free_images_iter,
            expr.args,  # pyright: ignore [reportGeneralTypeIssues]
        )
    )



import pytest

exports = (
    ('create_results',),
    ('fit',),
    ('read_results',),
    ('retrieve_models',),
    ('run_allometry',),
    ('run_amd',),
    ('run_covsearch',),
    ('run_iivsearch',),
    ('run_iovsearch',),
    ('run_modelsearch',),
    ('run_ruvsearch',),
    ('run_tool',),
)


@pytest.mark.parametrize(('tool',), exports)
def test_import_tools_all(tool):
    import pharmpy.tools as tools

    assert tool in getattr(tools, '__all__')


@pytest.mark.parametrize(('tool',), exports)
def test_import_tools_dir(tool):
    import pharmpy.tools as tools

    assert tool in dir(tools)


@pytest.mark.parametrize(('tool',), exports)
def test_import_tools_attr(tool):
    import pharmpy.tools as tools

    assert callable(getattr(tools, tool))


@pytest.mark.parametrize(('tool',), exports)
def test_import_tools_import(tool):
    mod = __import__('pharmpy.tools', globals(), locals(), [tool], 0)
    assert callable(getattr(mod, tool))



from markdown import markdown
import IPython


def markdown_html(string):
    return IPython.display.HTML(markdown(string))


def print_model_diff(model_ref, model_new):
    model_ref = model_ref.model_code.split('\n')
    model_new = model_new.model_code.split('\n')

    diffs = list(set(model_new) - set(model_ref))

    css_default = 'style=font-style:normal;font-size:80%;font-family:monospace;'

    # Hex codes:
    #    * Green:  #D6F3E8
    #    * Blue:   #DEF4FD
    #    * Yellow: #FEFFD2
    css_diff = f'{css_default}background-color:#DEF4FD'

    str_full = '<p style=line-height:120%>'

    for row in model_new:
        if row in diffs:
            str_full += f'<text {css_diff}>{row}</text><br>'
        else:
            str_full += f'<text {css_default}>{row}</text><br>'

    str_full += '</p>'

    display(markdown_html(str_full))
